from django.db import models


class CommonColumns(models.Model):
    
    class Meta:
        abstract = True
        
    f_binary = models.BinaryField(null=True)
    f_boolean = models.NullBooleanField(null=True)
    f_string = models.CharField(max_length=255, null=True)
    f_text = models.TextField(null=True)
    f_integer = models.IntegerField(null=True)
    f_float = models.FloatField(null=True)
    f_decimal = models.DecimalField(null=True,
                                    max_digits=18,
                                    decimal_places=6)
    f_date = models.DateField(null=True)
    f_time = models.TimeField(null=True)
    f_datetime = models.DateTimeField(null=True)


class ModelGeneralOne(CommonColumns):
    
    class Meta:
        db_table = 'model_general_one'
        
    #--- self relations ---
    
    # self one-to-one
    parent_1_1 = models.OneToOneField(
        'ModelGeneralOne',
        on_delete=models.CASCADE,
        related_name='child_1_1',
        null=True)
    
    # self one-to-many and many-to-one
    parent_x_1 = models.ForeignKey(
        'ModelGeneralOne',
        on_delete=models.CASCADE,
        related_name='child_1_x',
        null=True)
    
    # self many-to-many
    parent_x_x = models.ManyToManyField(
        'ModelGeneralOne',
        related_name='child_x_x')
    
    #--- relations with other ---
    
    # one-to-one auto configured on the other side
    
    # one-to-many auto configured on the other side with many-to-one
    
    # many-to-one
    two_x_1 = models.ForeignKey(
        'ModelGeneralTwo',
        on_delete=models.CASCADE,
        related_name='one_1_x',
        null=True)
    
    # many-to-many
    two_x_x = models.ManyToManyField(
        'ModelGeneralTwo',
        related_name='one_x_x')


class ModelGeneralTwo(CommonColumns):
    
    class Meta:
        db_table = 'model_general_two'
    
    #--- self relations ---
    
    # self one-to-one
    parent_1_1 = models.OneToOneField(
        'ModelGeneralTwo',
        on_delete=models.CASCADE,
        related_name='child_1_1',
        null=True)
    
    # self one-to-many and many-to-one
    parent_x_1 = models.ForeignKey(
        'ModelGeneralTwo',
        on_delete=models.CASCADE,
        related_name='child_1_x',
        null=True)
    
    # self many-to-many
    parent_x_x = models.ManyToManyField(
        'ModelGeneralTwo',
        related_name='child_x_x')
    
    #--- relations with other ---
    
    # one-to-one
    one_1_1 = models.OneToOneField(
        ModelGeneralOne,
        on_delete=models.CASCADE,
        related_name='two_1_1',
        null=True)

    # one-to-many auto configured on the other side with many-to-one
    
    # many-to-one
    one_x_1 = models.ForeignKey(
        ModelGeneralOne,
        on_delete=models.CASCADE,
        related_name='two_1_x',
        null=True)
    
    # many-to-many auto configured on the other side with many-to-many
    
    
    


""" Module that contains tests for adapters. The tests require models:

    - All models must have "f_binary", "f_boolean", "f_string", "f_text",
      "f_integer", "f_float", "f_date", "f_time", "f_datetime" columns with
      appropriate types.

    Models for basic adapter tests:

        - `ModelFieldTypes` with columns defined according to the rules:
            
            - "id" can be present or not (ignored).
            - Model fields (table columns) must be prefixed with it's type and
              underscore, for example: "integer_first". Types are just lower
              case versions of
              :py:class:`~save_to_db.adapters.utils.column_type.ColumnType`
              enumeration members.
            - Column names in database must be prefixed with 'new_'.
            - All types must be used at least once, except for `OTHER` type
              which must be not used at all.
    
    Models for basic relationship tests, must contains all types of
    relationships between different models and with self:
    
        - `ModelGeneralOne` and `ModelGeneralTwo`
          with next fields, relations and other definitions:
          
            - `ModelGeneralOne` must have "model_general_one" table name,
              `ModelGeneralTwo` must have "model_general_two" table name.
            - "id" primary key and any other fields used only as foreign keys.
            - both table must have next self referencing fields: 
                
                - "id" primary key
                - `parent_x_1` many-to-one relationship, `child_1_x` many-to-one
                  relationship on the other side.
                - `parent_1_1` one-to-one relationship, `child_1_1` one-to-one
                  relationship on the other side.
                - `parent_x_x` many-to-many relationship, `child_x_x`
                  many-to-many relationship on the other side.
            
            - `ModelGeneralOne.two_1_x` one-to-many relationship,
              `ModelGeneralTwo.one_x_1` many-to-one relationship
              on the other side.
            - `ModelGeneralOne.two_1_1` one-to-one relationship,
              `ModelGeneralTwo.one_1_1` one-to-one relationship
              on the other side.
            - `ModelGeneralOne.two_x_x` many-to-many
              relationship, `ModelGeneralTwo.one_x_x` many-to-many
              relationship on the other side.
            - `ModelGeneralOne.two_x_1` many-to-one relationship,
              `ModelGeneralTwo.one_1_x` one-to-many relationship
              on the other side.
              
            - all foreign keys must be "delete on cascade".
            
        - `ModelManyRefsOne` and `ModelManyRefsTwo`
          with next fields, relations and other definitions:
          
            - `ModelManyRefsOne` must have "model_many_refs_one" table name,
              `ModelManyRefsTwo` must have "model_many_refs_two" table name.
            - "id" primary key.
            
            - `ModelManyRefsOne.two_x_1_a` many-to-one relationship,
              `ModelManyRefsTwo.one_1_x_a` one-to-many relationship
              on the other side.
            - `ModelManyRefsOne.two_1_x_b` one-to-many relationship,
              `ModelManyRefsTwo.one_x_1_b` many-to-many relationship
              on the other side.
              
            - `ModelManyRefsOne.two_x_x_a` many-to-many relationship,
              `ModelManyRefsTwo.one_x_x_a` many-to-many relationship
              on the other side.
            - `ModelManyRefsOne.two_x_x_b` many-to-many relationship,
              `ModelManyRefsTwo.one_x_x_b` many-to-many relationship
              on the other side.
              
            - `ModelManyRefsOne.two_1_1_a` one-to-one relationship,
              `ModelManyRefsTwo.one_1_1_a` one-to-one relationship
              on the other side. Table one has foreign key to table two.
            - `ModelManyRefsOne.two_1_1_b` one-to-one relationship,
              `ModelManyRefsTwo.one_1_1_b` one-to-one relationship
              on the other side. Table two has foreign key to table one.
              
            - all foreign keys must be "delete on cascade".
    
    Models for creators and getters auto configuration should contain:
    
        - Models with single primary key and relationships using it:
        
            - many-to-one with required value.
            - many-to-one with not required value.
            
        - Models with composite primary key:
        
            - many-to-one with required value.
            - many-to-one with not required value.
            
        - Models null constraint:
            
            - Model with not null field.
        
        - Models with unique constraints:
            
            - Model with simple unique field.
            - Model with simple unique field that cannot be null.
            - Model with composite unique constraint.
            - Model with simple foreign key that has a unique constraint.
            - Model with simple not null foreign key that has a unique
              constraint.
            - Model with simple foreign key that has a unique constraint
              together with other column.
            - Model with composite foreign key that has a unique constraint.
        
        - Model with many-to-many-relationships.
        - Model with many-to-one not null relation with itself, one-to-one
          relationship with unique constraint with itself, one-to-one not null
          relationship with unique constraint with itself, many-to-many
          relationship with itself.
        
    And the models are:
    
        - `ModelConstraintsOne`:
            
            - `model_constraints_one` table name
            - `id` primary key
            - `f_text` cannot be null
            - `f_integer` unique
            - `f_string` not null and unique
            
        - `ModelConstraintsTwo`:
        
            - `model_constraints_two` table name
            - `id` primary key
            - `one_x_1_id` foreign key to `ModelConstraintsOne` with `one_x_1`
              relation and `two_1_x` on the other side
            - `three_x_x` relation with `ModelConstraintsThree` with `two_x_x`
              on the other side
            - `four_primary_one` and `four_primary_two` both not null foreign
              keys to `ModelConstraintsFour` with `four_x_1` relation and
              `two_1_x` on the other side
        
        - `ModelConstraintsThree`:
        
            - `model_constraints_three` table name
            - `id` primary key
            - `one_x_1_id` not null foreign key to `ModelConstraintsOne` with
              `one_x_1` relation and `three_1_x` on the other side
            - `two_x_x` relation with `ModelConstraintsTwo` and `three_x_x` on
              the other side
            - `four_primary_one` and `four_primary_two` foreign keys to
              `ModelConstraintsFour` that can be null, `four_x_1` relation with
              `three_1_x` on the other side
        
        - `ModelConstraintsFour`:
        
            - `model_constraints_four` table name
            - composite `primary_one` plus `primary_two` columns primary key
            - `f_integer` plus `f_string` unique constraint
            
        - `ModelConstraintsFive`:
        
            - `model_constraints_five` table name
            - `id` primary key
            - `one_1_1_id` unique foreign key to `ModelConstraintsOne` with
              `one_1_1` relation and `five_1_1` on the other side
            - `four_primary_one` plus `four_primary_two` unique foreign key
              to `ModelConstraintsFour` with `four_1_1` relation and `five_1_1`
              on the other side
        
        - `ModelConstraintsSix`:
        
            - `model_constraints_six` table name
            - `id` primary key
            - `five_1_1_id` not null foreign key to `ModelConstraintsFive` with
              `five_1_1` relation and `six_1_1` on the other side
            - `five_1_1_id` plus `f_integer` unique constraint
            
        - `ModelConstraintsSelf`:
        
            - `model_constraints_self` table name
            - `id` primary key
            - `parent_x_1_id` not null foreign key to itself with `parent_x_1`
              relation and `child_1_x` on the other side.
            - `first_parent_1_1_id` foreign key with unique constraint to
              itself with `first_parent_1_1` relation and `first_child_1_1` on
              the other side.
            - `parent_x_x` many-to-many relationship with itself with
              `child_x_x` on the other side.
            
    Models for testing reverse relations:
    
        Models here must include all possible kinds of setting up relationships,
        manual and automatic.

        - `ModelAutoReverseOne`:
            - `model_auto_reverse_one` table name
            - `two_b_1_1_id` foreign key to `ModelAutoReverseTwoB` with
              `two_b_1_1` with no relation on the other side.
            - `three_b_x_1_id` foreign key to `ModelAutoReverseTwoB` with
              `three_b_x_1` with no relation on the other side.
            - `four_b_x_x` to `ModelAutoReverseFourB` with no relation on the
               other side.
            - any other relations.
            - `f_string`
        - `ModelAutoReverseTwoA`:
            - `model_auto_reverse_two_a` table name
            - any relations.
        - `ModelAutoReverseTwoB`:
            - `model_auto_reverse_two_b` table name
            - any relations.
        - `ModelAutoReverseThreeA`:
            - `model_auto_reverse_three_a` table name
            - any relations.
        - `ModelAutoReverseThreeB`:
            - `model_auto_reverse_three_b` table name
            - any relations.
        - `ModelAutoReverseFourA`:
            - `model_auto_reverse_four_a` table name
            - any relations.
        - `ModelAutoReverseFourB`:
            - `model_auto_reverse_four_b` table name
            - any relations.
        
        All models must have `ITEM_RELATIONS` constant that contains info about
        relations like this:
        
            ITEM_RELATIONS = {
                key_1: (reverse_key_1, forword_relation_type_1,),
                ...
            }
            
    Models for testing merge policy:
        
        - `ModelOne`:
            - `two_1_1` foreign key to `ModelTwo` (reverse: `one_1_1`)
            - `two_x_1` foreign key to `ModelTwo` (reverse: `one_1_x`)
        - `ModelTwo`
            - `three_1_1` foreign key to `ModelTwo` (reverse: `two_1_1`)
            - `three_x_1` foreign key to `ModelTwo` (reverse: `two_1_x`)
        - `ModelThree`
            - Default fields only.
        - `ModelOneX`:
            - `two_x_1` foreign key to `ModelTwo` (reverse: `one_1_x`)
            - `f_integer` field.
            - unqiue together for both fields.
        - `ModelTwoX`
            - `three_x_1` foreign key to `ModelTwo` (reverse: `two_1_x`)
            - `f_integer` field.
            - unqiue together for both fields.
        - `ModelThreeX`
            - Default fields only.
            
        All models have `id` primary key. Default sort must be done by ID.
        
"""

import importlib
from unittest import TestCase

from save_to_db.adapters import ADAPTERS

# tuple of database adapter names to test
ADAPTER_NAMES = tuple(
    adapter_cls.__name__[:-len('Adapter')].lower()
    for adapter_cls in ADAPTERS
)


def init():
    """ Loads tests for all adapters. (Loading `tests` module from
    an adapter test module must trigger registration of tests)
    """
    for adapter_name in ADAPTER_NAMES:
        module_name = \
            'save_to_db.tests.test_adapters.project_{}.register_tests'.format(
                adapter_name)
        importlib.import_module(module_name)
        

registered_tests = []

def register_test(adapter_prefix, test_base, test_case, models,
                  persister):
    """ Registers a test case to be run on test command.
    
    :param adapter_prefix: A string prefix that will be used as part of the
        test class name.
    :param test_base: Test base class to be added as one of the base classes of
        a new test case. Must contain code to create and drop tables
        from database.
    :param test_case: The test case with tests.
    :param models: A dictionary of models that are needed for tests to run.
    :param persister: An instance of
        :py:class:`~save_to_db.core.persistor.Persistor` (`autocommit`
        must be `True`).
    """
    global registered_tests
    
    new_test_case_dict = {
        'persister': persister,
        'db_adapter': persister.db_adapter,
        'adapter_prefix': adapter_prefix,
    }
    new_test_case_dict.update(models)
    new_test_case = type('{}{}'.format(adapter_prefix, test_case.__name__),
                         (test_base, test_case),
                         new_test_case_dict)
    registered_tests.append(new_test_case)



import copy
from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestItemCopy(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        class ItemGeneralOne(Item):
            model_cls = cls.ModelGeneralOne
        cls.ItemGeneralOne = ItemGeneralOne
        
        class ItemGeneralTwo(Item):
            model_cls = cls.ModelGeneralTwo
        cls.ItemGeneralTwo = ItemGeneralTwo
    
    
    def test_copy(self):
        # --- single item ---
        item_one = self.ItemGeneralOne(f_integer=10)
        item_two = self.ItemGeneralTwo(f_integer=20)
        item_one['two_x_1'] = item_two
        item_one['two_1_x'].add(item_two)
        item_one['two_1_x__f_integer'] = 30
        
        item_one_copy = copy.copy(item_one)
        self.assertIsNot(item_one_copy, item_one)
        
        # related items must be the same
        self.assertIs(item_one_copy['two_x_1'], item_two)
        self.assertIs(item_one_copy['two_1_x'][0], item_two)
        
        self.assertEqual(item_one_copy['f_integer'], 10)
        self.assertEqual(item_one_copy['two_x_1']['f_integer'], 20)
        self.assertEqual(item_one_copy['two_1_x'][0]['f_integer'], 20)
        self.assertEqual(item_one_copy['two_1_x']['f_integer'], 30)
        
        # reverse relationships must keep updating
        second_item_two = self.ItemGeneralTwo(f_integer=200)
        item_one_copy['two_1_1'] = second_item_two
        self.assertIs(second_item_two['one_1_1'], item_one_copy)
        
        # --- bulk item ---
        bulk_one = self.ItemGeneralOne.Bulk(f_integer=10)
        item_one_in_bulk = self.ItemGeneralTwo()
        item_two_in_bulk_defaults = self.ItemGeneralTwo()
        
        bulk_one.add(item_one_in_bulk)
        bulk_one['two_x_1'] = item_two_in_bulk_defaults
        
        bulk_one_copy = copy.copy(bulk_one)
        self.assertIsNot(bulk_one_copy, bulk_one)
        self.assertEqual(bulk_one_copy['f_integer'], 10)
        
        # bulk attribute must be new but contain the same items
        self.assertIsNot(bulk_one_copy.bulk, bulk_one.bulk)
        self.assertEqual(bulk_one_copy.bulk, bulk_one.bulk)
        
        # related items must be the same ...
        
        # ... in bulk
        self.assertIs(bulk_one_copy[0], item_one_in_bulk)
        
        # ... in defaults
        self.assertIs(bulk_one['two_x_1'], item_two_in_bulk_defaults)
        
        # reverse relationships must keep updating
        item_two_container = self.ItemGeneralTwo(
            one_1_x=bulk_one_copy
        )
        second_item_one_in_bulk = self.ItemGeneralOne()
        bulk_one_copy.add(second_item_one_in_bulk)
        self.assertIs(second_item_one_in_bulk['two_x_1'], item_two_container)
    
    
    def test_deepcopy(self):
        # --- single item ---
        item_one = self.ItemGeneralOne(f_integer=10)
        item_two = self.ItemGeneralTwo(f_integer=20)
        item_one['two_x_1'] = item_two
        item_one['two_1_x'].add(item_two)
        item_one['two_1_x__f_integer'] = 30
        
        item_one_copy = copy.deepcopy(item_one)
        self.assertIsNot(item_one_copy, item_one)
        
        # related items must be new
        self.assertIsNot(item_one_copy['two_x_1'], item_two)
        self.assertIsNot(item_one_copy['two_1_x'][0], item_two)
        
        # reverse relationships mut be updated
        two_x_1 = item_one_copy['two_x_1']
        self.assertIs(two_x_1['one_1_x'], item_one_copy)
        
        self.assertEqual(item_one_copy['f_integer'], 10)
        self.assertEqual(item_one_copy['two_x_1']['f_integer'], 20)
        self.assertEqual(item_one_copy['two_1_x'][0]['f_integer'], 20)
        self.assertEqual(item_one_copy['two_1_x']['f_integer'], 30)
        
        return
        
        # reverse relationships must keep updating
        item_two = self.ItemGeneralTwo()
        item_one_copy['two_1_1'] = item_two
        self.assertIs(item_two['one_1_1'], item_one_copy)
        
        item_two = self.ItemGeneralTwo()
        item_one_copy['two_x_1'] = item_two
        self.assertEqual(len(item_two['one_1_x']), 1)
        self.assertIs(item_two['one_1_x'][0], item_one_copy)
        
        item_two = self.ItemGeneralTwo()
        item_one_copy['two_1_x'].add(item_two)
        self.assertIs(item_two['one_x_1'], item_one_copy)
        
        # --- bulk item ---
        bulk_one = self.ItemGeneralOne.Bulk(f_integer=10)
        item_one_in_bulk = self.ItemGeneralTwo()
        item_two_in_bulk_defaults = self.ItemGeneralTwo()
        
        bulk_one.add(item_one_in_bulk)
        bulk_one['two_x_1'] = item_two_in_bulk_defaults
        
        print()
        print('!!!')
        bulk_one_copy = copy.deepcopy(bulk_one)
        self.assertIsNot(bulk_one_copy, bulk_one)
        self.assertEqual(bulk_one_copy['f_integer'], 10)
        
        # copied items must be new ...
        
        # ... in bulk
        self.assertIsNot(bulk_one_copy[0], item_one_in_bulk)
        
        # ... in defaults
        self.assertIsNot(bulk_one['two_x_1'], item_two_in_bulk_defaults)
        
        pass


from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker


engine = create_engine('sqlite:///:memory:', echo=False)
Base = declarative_base()

DBSession = sessionmaker(bind=engine)
session = DBSession()


import sqlalchemy as sa

from ..config import Base



class ModelInvalidFieldNames(Base):
    __tablename__ = 'model_invalid_field_names'
    
    id = sa.Column(sa.Integer, primary_key=True)
    f__integer = sa.Column(sa.Integer)


""" This module contains exceptions for
:py:class:`~save_to_db.core.merge_policy.MergePolicy` class.
"""

class MergePolicyException(Exception):
    """ General exception for
    :py:class:`~save_to_db.core.merge_policy.MergePolicy` class.
    """

class ModelClsAlreadyRegistered(MergePolicyException):
    """ Raised when trying to add ORM model class to a merge policu twice.
    """
    
class UnknownRelationFieldNames(MergePolicyException):
    """ Raised when trying to setup merge policy using unknown relation
    relationship field name.
    """

class UnknownModelDefaultKey(MergePolicyException):
    """ Raised when trying to set defaults for merge policy using unkown
    default key.
    """


from django.apps import apps

from django.db import models, transaction
from django.db.models import Q

from django.db.models.fields.related import OneToOneField, OneToOneRel
from django.db.models.fields.related import ManyToManyField, ManyToManyRel
from django.db.models.fields.related import ManyToOneRel
from django.db.models.fields.related import ForeignKey

# from django.db.models.fields.related_descriptors import \
#     ReverseOneToOneDescriptor, ReverseManyToOneDescriptor

from .utils.adapter_base import AdapterBase
from .utils.column_type import ColumnType
from .utils.relation_type import RelationType


#--- start debug tool ----------------------------------------------------------
from django.db import connection
from functools import wraps

class _DebugQueries(object):
    
    def start(self):
        self.start_query_count = len(connection.queries)
        
    def end(self):
        print('+'*100)
        for i, q in enumerate(connection.queries[self.start_query_count:]):
            print('-'*80)
            print(i, q['sql'])
        print('*'*100, flush=True)
        
    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end()

dq = _DebugQueries()

def _dq_decorator(func):
    @wraps(func)
    def dq_decorated(self, *args, **kwargs):
        global dq
        with dq:
            return func(self, *args, **kwargs)
            
    return dq_decorated
#--- end debug tool ------------------------------------------------------------        


class DjangoAdapter(AdapterBase):
    """ An adapter for working with Django ORM.
    
    The `adapter_settings` is a dictionary with next values:
    
        - *using* is a database name (a key from `DATABASES` dictionary in the
          "settings.py" file of a Django project), if it isn't provided
          "default" database is used.
    
    .. note::
    
        If you are going to use transactions provided by this library,
        set `autocommit` for the database to `False`:
        
            .. code-block:: Python
        
                from django.db import transaction
                from save_to_db import Persister
                from save_to_db.adapters import DjangoAdapter
                
                persister = Persister(DjangoAdapter, adapter_settings={})
                
                # Next function takes a `using` argument which should be the
                # name of a database. If it isn’t provided, Django uses the
                # "default" database.
                transaction.set_autocommit(False, using='database_name')
                
                try:
                    persister.persist(save_to_db_item)
                    persister.commit()
                except:
                    persister.rollback()
        
        Alternativy, you can use Django transactions directly:
        
            .. code-block:: Python
            
                from django.db import transaction
                from save_to_db import Persister
                from save_to_db.adapters import DjangoAdapter
                
                persister = Persister(DjangoAdapter, adapter_settings={})
                
                with transaction.atomic():
                    persister.persist(save_to_db_item)
    """
    
    COMPOSITE_KEYS_SUPPORTED = False
    REVERSE_MODEL_AUTOUPDATE_SUPPORTED = False
    SAVE_MODEL_BEFORE_COMMIT = True
    
    
    def __init__(self, adapter_settings):
        super().__init__(adapter_settings)
        self.using = adapter_settings.get('using', 'default')
    
    
    #--- general methods -------------------------------------------------------
    
    @classmethod
    def is_usable(cls, model_cls):
        return issubclass(model_cls, models.Model)
    
    
    def commit(self):
        transaction.commit(using=self.using)


    def rollback(self):
        transaction.rollback(using=self.using)
    
    
    @classmethod
    def iter_fields(cls, model_cls):
        foreign_key_classes = (ForeignKey, ManyToOneRel,
                               OneToOneField, OneToOneRel,
                               ManyToManyField, ManyToManyRel)
        
        type_to_const = (
            (models.BinaryField, ColumnType.BINARY),
            (models.TextField, ColumnType.TEXT),
            (models.CharField, ColumnType.STRING),
            ((models.AutoField, models.IntegerField), ColumnType.INTEGER),
            ((models.BooleanField,
              models.NullBooleanField), ColumnType.BOOLEAN),
            (models.DateTimeField, ColumnType.DATETIME),
            (models.FloatField, ColumnType.FLOAT),
            (models.DecimalField, ColumnType.DECIMAL),
            (models.DateField, ColumnType.DATE),
            (models.TimeField, ColumnType.TIME),
            (models.DateTimeField, ColumnType.DATETIME),
        )
        for field in model_cls._meta.get_fields():
            if isinstance(field, foreign_key_classes):
                continue
            
            yield_type = ColumnType.OTHER
            for field_class, column_type in type_to_const:
                if isinstance(field, field_class):
                    yield_type = column_type
                    break
            yield field.name, yield_type
    
    
    @classmethod
    def iter_relations(cls, model_cls):
        foreign_key_classes = (ForeignKey, ManyToOneRel,
                               OneToOneField, OneToOneRel,
                               ManyToManyField, ManyToManyRel)
        
        type_to_const = [
            [(OneToOneField, OneToOneRel), RelationType.ONE_TO_ONE],
            [ForeignKey, RelationType.MANY_TO_ONE],
            [ManyToOneRel, RelationType.ONE_TO_MANY],
            [(ManyToManyField,
              ManyToManyRel), RelationType.MANY_TO_MANY],
        ]
        for field in model_cls._meta.get_fields():
            if not isinstance(field, foreign_key_classes):
                continue
            
            for field_class, column_type in type_to_const:
                if isinstance(field, field_class):
                    yield_type = column_type
                    break
            
            field_name = field.name
            if hasattr(field, 'get_accessor_name'):
                field_name = field.get_accessor_name()
            
            remote_field_name = None
            if not hasattr(field.remote_field, 'is_hidden') or \
                    not field.remote_field.is_hidden():
                remote_field_name = field.remote_field.name
                if hasattr(field.remote_field, 'get_accessor_name'):
                    remote_field_name = field.remote_field.get_accessor_name()
            
            yield field_name, field.related_model, yield_type, \
                remote_field_name
    
    
    @classmethod
    def iter_required_fields(cls, model_cls):
        for field in model_cls._meta.get_fields():
            if isinstance(field, (models.AutoField, ManyToManyField)):
                continue
            if not field.null and field.default is models.NOT_PROVIDED:
                yield field.name
    
    
    @classmethod
    def iter_unique_field_combinations(cls, model_cls):
        for unique_constraint_fields in model_cls._meta.unique_together:
            yield unique_constraint_fields
            
        for field in model_cls._meta.get_fields():
            if hasattr(field, 'field'):  # relation
                if isinstance(field, (OneToOneField,
                                      OneToOneRel)):
                    yield {field.name,}  # one-to-one always unique
                    continue
                field = field.field
            if field.unique:
                yield {field.name,}
    
    
    @classmethod
    def get_table_fullname(cls, model_cls):
        return model_cls._meta.db_table
    
    
    def get_model_cls_by_table_fullname(self, name):
        for models in apps.all_models.values():
            for model in models.values():
                if self.get_table_fullname(model) == name:
                    return model
    
    
    def iter_all_models(self):
        for models in apps.all_models.values():
            for model in models.values():
                yield model
    
    
    #--- methods for working with items ----------------------------------------
    

    def get(self, items_and_fkeys):

        #--- first getting item models from database ---------------------------
        if not items_and_fkeys:
            return []
        
        all_items_filters = Q()
        getters = items_and_fkeys[0][0].getters
        
        for item, fkeys in items_and_fkeys:
            one_item_filters = Q()
            for group in getters:
                group_filters = Q()
                skip_group_filters = False
                # in case related item was not in database
                for field_name in group:
                    if field_name not in item:
                        skip_group_filters = True
                        break
                    
                    if field_name in item.fields:
                        field_value = item[field_name]
                        group_filters &= Q(**{field_name: field_value})
                    elif field_name in item.relations:
                        related_models = fkeys.get(field_name)
                        if not related_models:
                            # failed to get or created related model before
                            skip_group_filters = True
                            break
                        
                        relation = item.relations[field_name]
                        if not relation['relation_type'].is_x_to_many():
                            group_filters &= Q(
                                **{field_name: related_models[0]})
                        else:
                            contains_any = Q()
                            for related_model in related_models:
                                contains_any |= Q(**{field_name: related_model})
                            group_filters &= contains_any
                
                if not skip_group_filters and len(group_filters):
                    one_item_filters |= group_filters
                
            if one_item_filters.children:
                all_items_filters |= one_item_filters
        
        if not len(all_items_filters):
            return []
        
        return item.model_cls.objects.db_manager(self.using) \
            .filter(all_items_filters).all()
    

    def delete(self, model):
        model.delete(using=self.using)
    

    def create_blank_model(self, model_cls):
        return model_cls()
    

    def add_related_models(self, model, fkey, related_models):
        getattr(model, fkey).add(*related_models)
                

    def clear_related_models(self, model, fkey):
        related = getattr(model, fkey)
        if hasattr(related, 'clear'): 
            getattr(model, fkey).clear()
        else:
            raise Exception('Cannot clear required field.')
        

    def related_x_to_many_exists(self, model, fkey):
        return getattr(model, fkey).exists()
    

    def related_x_to_many_contains(self, model, fkey, child_models):
        if not child_models:
            return []
        
        # no composite PK in Django
        pk_name = self.get_primary_key_names(model)[0]
        pk_list = [getattr(child_model, pk_name)
                   for child_model in child_models]
        contained_models = getattr(model, fkey).filter(**{
            '{}__in'.format(pk_name): pk_list
        })
        result = []
        for contained_model in contained_models:
            for child_model in child_models:
                if child_model == contained_model:
                    result.append(child_model)
                    break
                
        return result
    
    @classmethod
    def get_primary_key_names(cls, model_cls):
        return (model_cls._meta.pk.name,)
    

    def save_model(self, model):
        model.save(using=self.using)

    
    @classmethod
    def __get_select_and_keep_fielter(cls, selectors, keepers):
        total_select_filter = Q()
        for selectors_entry in selectors:
            select_filter_set = Q()
            for field_name, value in selectors_entry.items():
                select_filter_set &= Q(**{field_name: value})
            total_select_filter |= select_filter_set
        
        total_keep_filter = Q()
        for keepers_entry in keepers:
            keep_fielter_set = Q()
            for field_name, value in keepers_entry.items():
                keep_fielter_set |= ~Q(**{field_name: value})
            total_keep_filter &= keep_fielter_set
        
        if total_select_filter or total_keep_filter:
            return Q(total_select_filter, total_keep_filter)
        return None
    

    def execute_delete(self, model_cls, selectors, keepers):
        filters = self.__get_select_and_keep_fielter(selectors, keepers)
        if filters is None:
            return

        model_cls.objects.db_manager(self.using).filter(filters).delete()
    

    def execute_unref(self, parent, fkey, selectors, keepers):
        filters = self.__get_select_and_keep_fielter(selectors, keepers)
        if filters is None:
            return

        child_relation = getattr(parent, fkey)
        children = getattr(parent, fkey).filter(filters).all()
        child_relation.remove(*children)
    

    def get_related_x_to_many(self, model, fkey):
        return getattr(model, fkey).all()
    
    #--- methods for tests -----------------------------------------------------
    

    def get_all_models(self, model_cls, sort_key=None):
        models = list(model_cls.objects.db_manager(self.using).all())
        if sort_key:
            models.sort(key=sort_key)
        return models



from . import test_adapters
from .test_core import *
from .test_core_utils import *

# loading all adapter tests into current name space
test_adapters.init()
for registered_test in test_adapters.registered_tests:
    globals()[registered_test.__name__] = registered_test
registered_test = None  # removing reference in order not to load with `inspect`


import datetime

from save_to_db.core.scope import Scope
from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestPersist(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    ModelAutoReverseOne = None
    ModelAutoReverseTwoA = None
    ModelAutoReverseTwoB = None
    ModelAutoReverseThreeA = None
    ModelAutoReverseThreeB = None
    ModelAutoReverseFourA = None
    ModelAutoReverseFourB = None
    
    
    def test_forward_and_reverse_relations_filters_one(self):
        # referencing parent and two
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'parent_1_1'}, {'parent_x_1'}, {'parent_x_x'},
                        {'two_1_1'}, {'two_1_x'}, {'two_x_1'}, {'two_x_x'},
                        {'f_integer'}]
            getters = [{'parent_1_1'}, {'parent_x_1'}, 
                       {'two_1_1'}, {'two_x_1'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            
        persister = self.persister
        
        # parent_1_1
        item_parent_1_1 = ItemGeneralOne(f_string='parent_1_1',
                                         f_integer='101')
        item_child_1_1 = ItemGeneralOne(f_string='child_1_1',
                                        parent_1_1=item_parent_1_1)
        persister.persist(item_child_1_1)
        models = self.get_all_models(self.ModelGeneralOne)
        model = [m for m in models if m.f_string=='child_1_1'][0]
        self.assertEqual(model.parent_1_1.f_string, 'parent_1_1')
        model = [m for m in models if m.f_string=='parent_1_1'][0]
        self.assertEqual(model.child_1_1.f_string, 'child_1_1')
        
        # parent_x_1
        item_parent_x_1 = ItemGeneralOne(f_string='parent_x_1',
                                         f_integer='901')
        item_child_1_x = ItemGeneralOne(f_string='child_1_x',
                                        parent_x_1=item_parent_x_1)
        persister.persist(item_child_1_x)
        models = self.get_all_models(self.ModelGeneralOne)
        model = [m for m in models if m.f_string=='child_1_x'][0]
        self.assertEqual(model.parent_x_1.f_string, 'parent_x_1')
        model = [m for m in models if m.f_string=='parent_x_1'][0]
        child_1_x = self.get_related_x_to_many(model, 'child_1_x')
        self.assertEqual(len(child_1_x), 1)
        self.assertEqual(child_1_x[0].f_string, 'child_1_x')
        
        # two_1_1
        item_two_1_1 = ItemGeneralTwo(f_string='two_1_1',
                                      f_integer='101')
        item_one_1_1 = ItemGeneralOne(f_string='one_1_1',
                                      two_1_1=item_two_1_1)
        persister.persist(item_one_1_1)
        
        models = self.get_all_models(self.ModelGeneralOne)
        model = [m for m in models if m.f_string=='one_1_1'][0]
        self.assertEqual(model.two_1_1.f_string, 'two_1_1')
        models = self.get_all_models(self.ModelGeneralTwo)
        model = [m for m in models if m.f_string=='two_1_1'][0]
        self.assertEqual(model.one_1_1.f_string, 'one_1_1')
        
        # two_x_1
        item_two_x_1 = ItemGeneralTwo(f_string='two_x_1',
                                      f_integer='901')
        item_one_1_x = ItemGeneralOne(f_string='one_1_x',
                                      two_x_1=item_two_x_1)
        persister.persist(item_one_1_x)
        models = self.get_all_models(self.ModelGeneralOne)
        model = [m for m in models if m.f_string=='one_1_x'][0]
        self.assertEqual(model.two_x_1.f_string, 'two_x_1') 
        models = self.get_all_models(self.ModelGeneralTwo)
        model = [m for m in models if m.f_string=='two_x_1'][0]
        one_1_x = self.get_related_x_to_many(model, 'one_1_x')
        self.assertEqual(len(one_1_x), 1)
        self.assertEqual(one_1_x[0].f_string, 'one_1_x')
    
    
    def test_forward_and_reverse_relations_filters_two(self):
        # referencing child and one
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'child_1_1'}, {'child_x_x'},
                        {'f_integer'}]
            getters = [{'child_1_1'},
                       {'two_1_1'}, {'two_x_1'},
                       {'f_string'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'one_1_1'}, {'one_1_x'}, {'one_x_1'}, {'one_x_x'},
                        {'f_integer'}]
            getters = [{'one_1_1'}, {'one_x_1'},
                       {'f_integer'}]
        
        persister = self.persister
        
        # child_1_1
        item_child_1_1 = ItemGeneralOne(f_string='child_1_1',
                                        f_integer='101')
        item_parent_1_1 = ItemGeneralOne(f_string='parent_1_1',
                                         child_1_1=item_child_1_1)
        persister.persist(item_parent_1_1)
        models = self.get_all_models(self.ModelGeneralOne)
        model = [m for m in models if m.f_string=='parent_1_1'][0]
        self.assertEqual(model.child_1_1.f_string, 'child_1_1')
        model = [m for m in models if m.f_string=='child_1_1'][0]
        self.assertEqual(model.parent_1_1.f_string, 'parent_1_1')
        
        # one_1_1
        item_one_1_1 = ItemGeneralOne(f_string='one_1_1',
                                      f_integer='1101')
        item_two_1_1 = ItemGeneralTwo(f_string='two_1_1',
                                      one_1_1=item_one_1_1)
        persister.persist(item_two_1_1)
        models = self.get_all_models(self.ModelGeneralTwo)
        model = [m for m in models if m.f_string=='two_1_1'][0]
        self.assertEqual(model.one_1_1.f_string, 'one_1_1')
        models = self.get_all_models(self.ModelGeneralOne)
        model = [m for m in models if m.f_string=='one_1_1'][0]
        self.assertEqual(model.two_1_1.f_string, 'two_1_1')
        
        # one_x_1
        item_one_x_1 = ItemGeneralOne(f_string='one_x_1',
                                      f_integer='1901')
        item_two_1_x = ItemGeneralTwo(f_string='two_1_x',
                                      one_x_1=item_one_x_1)
        persister.persist(item_two_1_x)
        models = self.get_all_models(self.ModelGeneralTwo)
        model = [m for m in models if m.f_string=='two_1_x'][0]
        self.assertEqual(model.one_x_1.f_string, 'one_x_1') 
        models = self.get_all_models(self.ModelGeneralOne)
        model = [m for m in models if m.f_string=='one_x_1'][0]
        two_1_x = self.get_related_x_to_many(model, 'two_1_x')
        self.assertEqual(len(two_1_x), 1)
        self.assertEqual(two_1_x[0].f_string, 'two_1_x')
    
    
    def test_forward_and_reverse_relations_persist(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        persister = self.persister
        
        #--- from item_one ---
        item_one = ItemGeneralOne()
        item_one['two_1_1'] = ItemGeneralTwo(f_integer='101')
        item_one['two_x_1'] = ItemGeneralTwo(f_integer='901')
        item_one['two_x_x'].gen(f_integer='909')
        
        persister.persist(item_one)
        models = self.get_all_models(self.ModelGeneralOne)
        
        self.assertEqual(len(models), 1)
        model = models[0]
        self.assertEqual(model.two_1_1.f_integer, 101)
        self.assertEqual(model.two_x_1.f_integer, 901)
        two_x_x = self.get_related_x_to_many(model, 'two_x_x')
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(two_x_x[0].f_integer, 909)
        
        #--- from item two ---
        item_two = ItemGeneralTwo(f_string='second item')
        item_two['one_1_1'] = ItemGeneralOne(f_integer='2101')
        item_two['one_x_1'] = ItemGeneralOne(f_integer='2901')
        item_two['one_x_x'].gen(f_integer='2909')
        
        persister.persist(item_two)
        models = self.get_all_models(self.ModelGeneralTwo)
        
        #self.assertEqual(len(models), 4)  # 3 before and 1 now
        model = [m for m in models if m.f_string=='second item'][0]
        self.assertEqual(model.one_1_1.f_integer, 2101)
        self.assertEqual(model.one_x_1.f_integer, 2901)
        one_x_x = self.get_related_x_to_many(model, 'one_x_x')
        self.assertEqual(len(one_x_x), 1)
        self.assertEqual(one_x_x[0].f_integer, 2909)
        
        #--- from self to self ---
        item_one = ItemGeneralOne(f_string='third item')
        # parent
        item_one['parent_1_1'] = ItemGeneralOne(f_integer='3101')
        item_one['parent_x_1'] = ItemGeneralOne(f_integer='3901')
        item_one['parent_x_x'].gen(f_integer='3909')
        # child
        item_one['child_1_1'] = ItemGeneralOne(f_integer='4101')
        item_one['child_1_x'].gen(f_integer='4109')
        item_one['child_x_x'].gen(f_integer='4909')
        
        persister.persist(item_one)
        models = self.get_all_models(self.ModelGeneralOne)
        model = [m for m in models if m.f_string=='third item'][0]
        
        # parent
        self.assertEqual(model.parent_1_1.f_integer, 3101)
        self.assertEqual(model.parent_x_1.f_integer, 3901)
        parent_x_x = self.get_related_x_to_many(model, 'parent_x_x')
        self.assertEqual(len(parent_x_x), 1)
        self.assertEqual(parent_x_x[0].f_integer, 3909)
        
        # child
        self.assertEqual(model.child_1_1.f_integer, 4101)
        child_1_x = self.get_related_x_to_many(model, 'child_1_x')
        self.assertEqual(len(child_1_x), 1)
        self.assertEqual(child_1_x[0].f_integer, 4109)
        child_x_x = self.get_related_x_to_many(model, 'child_x_x')
        self.assertEqual(len(child_x_x), 1)
        self.assertEqual(child_x_x[0].f_integer, 4909)
    
    
    def test_persist_bulk_item(self):
        class ItemGeneralOne(Item):
            allow_merge_items = True
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float'}]
        
        class ItemGeneralTwo(Item):
            allow_merge_items = True
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float'}]
        
        persister = self.persister
        
        # generating 10 simple items
        bulk_one = ItemGeneralOne.Bulk()
        for i in range(10):
            bulk_one.gen(id=i, f_integer=i+100, f_float=i+1000)
        
        previous_bulk = bulk_one.bulk[:]
        items, model_lists = persister.persist(bulk_one)
        self.assertEqual(bulk_one.bulk, previous_bulk)
        
        self.assertEqual(len(items), 10)
        self.assertEqual(len(items), len(model_lists))
        models = [model_entries[0] for model_entries in model_lists
                  if self.assertEqual(len(model_entries), 1) is None]
        models.sort(key=lambda model: model.id)
        loaded_models = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(models), 10)
        for i, model in enumerate(models):
            self.assertIn(model, loaded_models)
            self.assertEqual(model.id, i)
            self.assertEqual(model.f_integer, i+100)
            self.assertEqual(model.f_float, i+1000)
            self.assertIs(model.f_boolean, None)  # for future tests
        
        # generating 10 items that has previous 5 and new 5 items as relations
        bulk_two = ItemGeneralTwo.Bulk()
        for i in range(10):
            item = bulk_two.gen(id=i, f_integer=i+200, f_float=i+2000)
            if i < 5:
                # last five item ones will have twos with IDs from 0 to 4
                # (all `i` values), item twos from 5 to 9
                item['one_x_x'].add(*bulk_one[5:])
                item['one_x_x__f_boolean'] = 'true'
                
                if i == 1:
                    # items are from previously saved bulk
                    # item one at index zero has ID = 0,
                    # item two has ID = 1 (current `i`, current item)
                    item['one_1_1'] = items[0]
                    item['one_1_x'].add(items[0])
            else:
                for j in range(5, 10):
                    # generated items are the same for all iterations,
                    # item twos from 5 to 9 will have item ones with IDs
                    # from 600 to 604, item ones on the other side will have
                    # item twos with IDs from 5 to 9 (all `i` values)
                    item['one_x_x'].gen(id=j+600-5, f_integer=j+200-5,
                                        f_float=j+2000-5)
        
        items, model_lists = persister.persist(bulk_two)
        self.assertEqual(len(items), 10)
        self.assertEqual(len(items), len(model_lists))
        models = [model_entries[0] for model_entries in model_lists
                  if self.assertEqual(len(model_entries), 1) is None]
        
        # checking item twos ---------------------------------------------------
        
        # 5 items were not changed, 5 items were updated, 5 items were created
        all_models = self.get_all_models(self.ModelGeneralOne)
        all_models.sort(key=lambda model: model.id)
        self.assertEqual(len(all_models), 15)
        
        # checking fields (using IDs for relations)
        expected_values = [
            # first 5 items that were not changed
            dict(id=0, f_integer=100, f_float=1000.0, f_boolean=None,
                 two_1_1=1, two_x_1=1),
            dict(id=1, f_integer=101, f_float=1001.0, f_boolean=None),
            dict(id=2, f_integer=102, f_float=1002.0, f_boolean=None),
            dict(id=3, f_integer=103, f_float=1003.0, f_boolean=None),
            dict(id=4, f_integer=104, f_float=1004.0, f_boolean=None),
            # 5 items that were updated
            dict(id=5, f_integer=105, f_float=1005.0, f_boolean=True,
                 two_x_x=[0,1,2,3,4]),
            dict(id=6, f_integer=106, f_float=1006.0, f_boolean=True,
                 two_x_x=[0,1,2,3,4]),
            dict(id=7, f_integer=107, f_float=1007.0, f_boolean=True,
                 two_x_x=[0,1,2,3,4]),
            dict(id=8, f_integer=108, f_float=1008.0, f_boolean=True,
                 two_x_x=[0,1,2,3,4]),
            dict(id=9, f_integer=109, f_float=1009.0, f_boolean=True,
                 two_x_x=[0,1,2,3,4]),
            # 5 items that were created
            dict(id=600, f_integer=200, f_float=2000.0, f_boolean=None,
                 two_x_x=[5,6,7,8,9]),
            dict(id=601, f_integer=201, f_float=2001.0, f_boolean=None,
                 two_x_x=[5,6,7,8,9]),
            dict(id=602, f_integer=202, f_float=2002.0, f_boolean=None,
                 two_x_x=[5,6,7,8,9]),
            dict(id=603, f_integer=203, f_float=2003.0, f_boolean=None,
                 two_x_x=[5,6,7,8,9]),
            dict(id=604, f_integer=204, f_float=2004.0, f_boolean=None,
                 two_x_x=[5,6,7,8,9]),
        ]
        relations = {
            fname: direction
            for fname, _, direction, _ in persister.db_adapter.iter_relations(
                all_models[0].__class__)
        }
        for i, (expected, model) in enumerate(zip(expected_values, all_models)):
            for key, value in expected.items():
                if key not in relations:
                    model_value = getattr(model, key)
                    err_msg = '{}: {} = {}'.format(i, key, model_value)
                    if key != 'f_boolean':
                        self.assertEqual(model_value, value, err_msg)
                    else:
                        self.assertIs(model_value, value, err_msg)
                else:
                    if relations[key].is_x_to_many():
                        model_value = self.get_related_x_to_many(model, key)
                        ids = list(m.id for m in model_value)
                        ids.sort()
                        self.assertEqual(
                            ids, value,
                            '{}: {} = {}'.format(i, key, ids))
                    else:
                        model_value = getattr(model, key)
                        self.assertEqual(
                            model_value.id, value,
                            '{}: {} = {}'.format(i, key, model_value.id))
        
        # checking item ones ---------------------------------------------------
        
        # 10 items were created, other creations or no changes were made
        # (5 items were reused, 5 items were created)
        all_models = self.get_all_models(self.ModelGeneralTwo)
        all_models.sort(key=lambda model: model.id)
        self.assertEqual(len(all_models), 10)
        
        expected_values = [
            dict(id=0, f_integer=200, f_float=2000.0,
                 one_x_x=[5,6,7,8,9]),
            dict(id=1, f_integer=201, f_float=2001.0,
                 one_x_x=[5,6,7,8,9],
                 one_1_1=0, one_1_x=[0]),
            dict(id=2, f_integer=202, f_float=2002.0,
                 one_x_x=[5,6,7,8,9]),
            dict(id=3, f_integer=203, f_float=2003.0,
                 one_x_x=[5,6,7,8,9]),
            dict(id=4, f_integer=204, f_float=2004.0,
                 one_x_x=[5,6,7,8,9]),
            
            dict(id=5, f_integer=205, f_float=2005.0,
                 one_x_x=[600,601,602,603,604]),
            dict(id=6, f_integer=206, f_float=2006.0,
                 one_x_x=[600,601,602,603,604]),
            dict(id=7, f_integer=207, f_float=2007.0,
                 one_x_x=[600,601,602,603,604]),
            dict(id=8, f_integer=208, f_float=2008.0,
                 one_x_x=[600,601,602,603,604]),
            dict(id=9, f_integer=209, f_float=2009.0,
                 one_x_x=[600,601,602,603,604]),
        ]
        relations = {
            fname: direction
            for fname, _, direction, _ in persister.db_adapter.iter_relations(
                all_models[0].__class__)
        }
        for i, (expected, model) in enumerate(zip(expected_values, all_models)):
            for key, value in expected.items():
                if key not in relations:
                    model_value = getattr(model, key)
                    err_msg = '{}: {} = {}'.format(i, key, model_value)
                    self.assertEqual(model_value, value, err_msg)
                else:
                    if relations[key].is_x_to_many():
                        model_value = self.get_related_x_to_many(model, key)
                        ids = list(m.id for m in model_value)
                        ids.sort()
                        self.assertEqual(
                            ids, value,
                            '{}: {} = {}'.format(i, key, ids))
                    else:
                        model_value = getattr(model, key)
                        self.assertEqual(
                            model_value.id, value,
                            '{}: {} = {}'.format(i, key, model_value.id))
    
    
    def test_persist_no_reverse_relations(self):
        class ItemAutoReverseOne(Item):
            model_cls = self.ModelAutoReverseOne
        
        class ItemAutoReverseTwoA(Item):
            model_cls = self.ModelAutoReverseTwoA
        
        class ItemAutoReverseTwoB(Item):
            model_cls = self.ModelAutoReverseTwoB
        
        class ItemAutoReverseThreeA(Item):
            model_cls = self.ModelAutoReverseThreeA
        
        class ItemAutoReverseThreeB(Item):
            model_cls = self.ModelAutoReverseThreeB
        
        class ItemAutoReverseFourA(Item):
            model_cls = self.ModelAutoReverseFourA
        
        class ItemAutoReverseFourB(Item):
            model_cls = self.ModelAutoReverseFourB

        
        item = ItemAutoReverseOne(f_string='one')
        
        item['two_b_1_1'] = ItemAutoReverseTwoB(f_string='two_b_1_1')
        item['three_b_x_1'] = ItemAutoReverseThreeB(f_string='three_b_x_1')
        item['four_b_x_x'].gen(f_string='four_b_x_x_first')
        item['four_b_x_x'].gen(f_string='four_b_x_x_second')
        
        persister = self.persister
        _, models_one_list = persister.persist(item)
        self.assertTrue(len(models_one_list), 1)
        models_one = models_one_list[0]
        self.assertTrue(len(models_one), 1)
        model_one_returned = models_one[0]
        
        # one
        models_one = self.get_all_models(self.ModelAutoReverseOne)
        self.assertEqual(len(models_one), 1)
        model_one = models_one[0]
        
        self.assertEqual(model_one, model_one_returned)
        
        # two
        models_two = self.get_all_models(self.ModelAutoReverseTwoB)
        self.assertEqual(len(models_two), 1)
        model_two = models_two[0]
        
        self.assertEqual(model_one.two_b_1_1, model_two)
        
        # three
        models_three = self.get_all_models(self.ModelAutoReverseThreeB)
        self.assertEqual(len(models_three), 1)
        model_three = models_three[0]
        
        self.assertEqual(model_one.three_b_x_1, model_three)
        
        # four
        sort_func = lambda model: model.f_string
        
        models_four = self.get_all_models(self.ModelAutoReverseFourB,
                                          sort_key=sort_func)
        self.assertEqual(len(models_four), 2)
        
        related_models_four = self.get_related_x_to_many(model_one,
                                                         'four_b_x_x',
                                                         sort_key=sort_func)
        self.assertEqual(len(related_models_four), 2)
        for model_four, related_model_for in zip(models_four,
                                                 related_models_four):
            self.assertEqual(model_four, related_model_for)
    
    
    def test_persist_single_item_no_related(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_float'}, {'f_string', 'f_text'},
                       {'f_date'}]
        
        # need this for `ItemGeneralOne` (related field)
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
                
        persister = self.persister
        
        # creating and updating first model ------------------------------------
        
        # cannot create or get
        item = ItemGeneralOne(f_string='str-one')
        items, model_lists = persister.persist(item)
        self.assertFalse(items)
        self.assertFalse(model_lists)
        
        # can get but not create and not in database
        item['f_text'] = 'text-one'
        items, model_lists = persister.persist(item)
        self.assertFalse(items)
        self.assertFalse(model_lists)
        
        # can create
        item['f_integer'] = None  # `None` can create
        items, model_lists = persister.persist(item)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(items), len(model_lists))
        models = model_lists[0]
        self.assertEqual(len(models), 1)
        self.assertIs(items[0], item)
        
        self.assertEqual(models[0].f_string, 'str-one')
        self.assertEqual(models[0].f_text, 'text-one')
        self.assertEqual(models[0].f_integer, None)
        self.assertEqual(models[0].f_date, None)
        
        # can update (`{'f_string', 'f_text'}` are the getters)
        
        # `f_date=None` will be used as getter later
        item['f_date'] = '2010-10-10'
        
        item['f_integer'] = '100'
        items, model_lists = persister.persist(item)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(items), len(model_lists))
        models = model_lists[0]
        self.assertEqual(len(models), 1)
        self.assertIs(items[0], item)
        
        # checking that model got the same values
        self.assertEqual(models[0].f_string, 'str-one')
        self.assertEqual(models[0].f_text, 'text-one')
        self.assertEqual(models[0].f_integer, 100)
        self.assertEqual(models[0].f_date, datetime.date(2010, 10, 10))
        
        first_model = models[0]
        
        # creating second model ------------------------------------------------
        
        item = ItemGeneralOne(f_integer='200', f_string='str-one')
        items, model_lists = persister.persist(item)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(items), len(model_lists))
        models = model_lists[0]
        self.assertEqual(len(models), 1)
        self.assertIs(items[0], item)
        
        second_model = models[0]
        self.assertNotEqual(models[0], first_model)
        self.assertEqual(models[0].f_string, 'str-one')
        self.assertEqual(models[0].f_integer, 200)
        
        # updating first and second model --------------------------------------
        
        # updating second model
        item = ItemGeneralOne(f_date=None)  # `None` value can be used
        items, model_lists = persister.persist(item)
        self.assertEqual(len(models), 1)
        self.assertIs(items[0], item)
        models = model_lists[0]
        self.assertEqual(len(models), 1)
        self.assertIs(items[0], item)
        
        self.assertEqual(models[0], second_model)
        self.assertEqual(models[0].f_integer, 200)
        self.assertEqual(models[0].f_string, 'str-one')
        self.assertEqual(models[0].f_text, None)
        self.assertIs(models[0].f_boolean, None)
        self.assertIs(models[0].f_date, None)
        
        # updating second model
        self.assertIs(first_model.f_boolean, None)  # will be `True` later
        item['f_string'] = 'str-one'
        item['f_text'] = 'text-one'
        item['f_boolean'] = True
        item['f_date'] = '2015-10-10'
        
        items, model_lists = persister.persist(item)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(items), len(model_lists))
        models = model_lists[0]
        self.assertEqual(len(models), 1)
        self.assertIs(items[0], item)
        
        self.assertEqual(models[0], first_model)
        self.assertEqual(models[0].f_integer, 100)
        self.assertEqual(models[0].f_string, 'str-one')
        self.assertEqual(models[0].f_text, 'text-one')
        self.assertIs(models[0].f_boolean, True)
        self.assertEqual(models[0].f_date, datetime.date(2015, 10, 10))

        # return boolean to `None`
        item['f_boolean'] = None
        items, (models,) = persister.persist(item)
        self.assertEqual(models[0], first_model)
        self.assertIs(models[0].f_boolean, None)
    
    
    def test_persist_single_item_with_single_related(self):
        model_one_cls = self.ModelGeneralOne
        model_two_cls = self.ModelGeneralTwo
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer', 'two_1_1'}]
            getters = [{'f_integer', 'two_1_1'},
                       {'f_float'},
                       {'f_string', 'f_text'},
                       {'f_date'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        persister = self.persister
        
        # cannot create any items
        # 'two_1_1' is not present 
        item_one = ItemGeneralOne(f_integer='100')
        items, model_lists = persister.persist(item_one)
        self.assertEqual(len(items), 0)
        self.assertEqual(len(items), len(model_lists))
        
        # 'two_1_1' cannot be created
        item_two = ItemGeneralTwo(f_float='20.20')
        items, model_lists = persister.persist(item_two)
        self.assertEqual(len(items), 0)
        self.assertEqual(len(items), len(model_lists))
        
        item_one['two_1_1'] = item_two
        items, model_lists = persister.persist(item_one)
        self.assertEqual(len(items), 0)
        self.assertEqual(len(items), len(model_lists))
        
        self.assertEqual(len(self.get_all_models(model_one_cls)), 0)
        self.assertEqual(len(self.get_all_models(model_two_cls)), 0)
        
        # 'two_1_1' can be created
        item_two['f_integer'] = '200'
        del item_one['f_integer']  # won't be persisted
        items, model_lists = persister.persist(item_one)
        self.assertEqual(len(items), 0)
        self.assertEqual(len(items), len(model_lists))
        
        self.assertEqual(len(self.get_all_models(model_one_cls)), 0)
        # but `item_two` was persisted
        models_two = self.get_all_models(model_two_cls)
        self.assertEqual(len(models_two), 1)
        self.assertEqual(models_two[0].f_integer, 200)
        self.assertEqual(models_two[0].f_float, 20.2)
        
        # persisting item_one with updating existing model
        self.assertIs(models_two[0].f_boolean, None)
        item_one['f_integer'] = '100'
        item_two['f_boolean'] = True
        
        items, model_lists = persister.persist(item_one)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(items), len(model_lists))
        self.assertEqual(len(model_lists), 1)
        models = model_lists[0]
        self.assertEqual(len(models), 1)
        self.assertIs(items[0], item_one)
        
        self.assertEqual(models[0].f_integer, 100)
        self.assertIsNotNone(models[0].two_1_1)
        self.assertIsNotNone(models[0].two_1_1.f_integer, 200)
        self.assertIsNotNone(models[0].two_1_1.f_float, 20.2)
        self.assertIs(models[0].two_1_1.f_boolean, True)
        
        self.assertEqual(len(self.get_all_models(model_one_cls)), 1)
        self.assertEqual(len(self.get_all_models(model_two_cls)), 1)
        
        # persisting using related item as filter
        del item_two['one_1_1']  # old item
        item_one = ItemGeneralOne(f_integer='100', two_1_1=item_two)
        item_one['f_boolean'] = False
        
        items, model_lists = persister.persist(item_one)
        
        self.assertEqual(len(items), 1)
        self.assertEqual(len(items), len(model_lists))
        self.assertEqual(len(model_lists), 1)
        models = model_lists[0]
        self.assertEqual(len(models), 1)
        self.assertIs(items[0], item_one)
        self.assertIs(models[0].f_boolean, False)
        
        self.assertEqual(len(self.get_all_models(model_one_cls)), 1)
        self.assertEqual(len(self.get_all_models(model_two_cls)), 1)
    
    
    def test_persist_single_item_with_bulk_related(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'id'}]
            getters = [{'id'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'id'}]
            getters = [{'id'}]
        
        persister = self.persister
        
        item_one = ItemGeneralOne(id='5', f_float='50.50')
        item_one(two_1_1=ItemGeneralTwo(id='10', f_string='1-to-1'),
                 two_x_1=ItemGeneralTwo(id='20', f_string='X-to-1'))
        bulk = item_one['two_1_x']
        bulk.gen(id='30', f_string='1-to-x first')
        bulk.gen(id='40', f_string='1-to-x second')
        bulk = item_one['two_x_x']
        bulk.gen(id='50', f_string='x-to-x first')
        bulk.gen(id='60', f_string='x-to-x second')
        
        items, model_lists = persister.persist(item_one)
        
        self.assertEqual(len(items), 1)
        self.assertEqual(len(items), len(model_lists))
        models = model_lists[0]
        self.assertEqual(len(models), 1)
        self.assertIs(items[0], item_one)
        
        # checking model data
        model = models[0]
        self.assertEqual(model.id, 5)
        self.assertEqual(model.f_float, 50.5)
        
        self.assertIsNotNone(model.two_1_1)
        self.assertEqual(model.two_1_1.id, 10)
        self.assertEqual(model.two_1_1.f_string, '1-to-1')
        
        two_1_x = self.get_related_x_to_many(model, 'two_1_x')
        two_1_x.sort(key=lambda model: model.id)
        self.assertEqual(len(two_1_x), 2)
        self.assertEqual(two_1_x[0].id, 30)
        self.assertEqual(two_1_x[0].f_string, '1-to-x first')
        self.assertEqual(two_1_x[1].id, 40)
        self.assertEqual(two_1_x[1].f_string, '1-to-x second')
        
        two_x_x = self.get_related_x_to_many(model, 'two_x_x')
        two_x_x.sort(key=lambda model: model.id)
        self.assertEqual(len(two_x_x), 2)
        self.assertEqual(two_x_x[0].id, 50)
        self.assertEqual(two_x_x[0].f_string, 'x-to-x first')
        self.assertEqual(two_x_x[1].id, 60)
        self.assertEqual(two_x_x[1].f_string, 'x-to-x second')
    
    
    def test_persist_single_item_with_x_to_many_related(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_string'}]
            getters = [{'two_1_x'}, {'f_string'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_string'}]
            getters = [{'f_string'}]
        
        persister = self.persister
        
        # creating first set of items
        item_one_bulk = ItemGeneralOne.Bulk()
        for (item_name, known_names) in [['one', ['first', 'one',]],
                                         ['two', ['second', 'two']]]:
            item_one = item_one_bulk.gen(f_string=item_name)
            for known_name in known_names:
                item_one['two_1_x'].gen(f_string=known_name)
            
        persister.persist(item_one_bulk)
        
        # creating second set of items
        # (`item_one` instance with name "one" must match 
        #  item with name "one_more" through "first" known name)
        item_one_bulk = ItemGeneralOne.Bulk()
        for (item_name, known_names) in [['one_more', ['first', 'one_more',]],
                                         ['three', ['third', 'three']]]:
            item_one = item_one_bulk.gen(f_string=item_name)
            for known_name in known_names:
                item_one['two_1_x'].gen(f_string=known_name)
            
        persister.persist(item_one_bulk)
        
        # updating two models (testing mathing x-to-many)
        item_one_bulk = ItemGeneralOne.Bulk()
        for (item_name, known_names) in [['two_more', ['second', 'two',
                                                       'two_more',]],
                                         ['three_more', ['third', 'three',
                                                         'three_more']]]:
            item_one = item_one_bulk.gen(f_string=item_name)
            for known_name in known_names:
                item_one['two_1_x'].gen(f_string=known_name)
        
        persister.persist(item_one_bulk)
        
        # checking the models
        sort_func = lambda model: model.f_string
        models_one = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=sort_func)
        self.assertEqual(len(models_one), 3)
        
        expected = [['one_more', ['first', 'one', 'one_more']],
                    ['two_more', ['second', 'two', 'two_more']],
                    ['three_more', ['third', 'three', 'three_more']]]
        expected.sort(key=lambda entry: entry[0])
        
        for i, (expected_name, expected_known_names) in enumerate(expected):
            
            model_one = models_one[i]
            self.assertEqual(model_one.f_string, expected_name)
            
            models_two = self.get_related_x_to_many(model_one, 'two_1_x',
                                                    sort_func)
            actual_known_names = [model_two.f_string
                                  for model_two in models_two]
            known_names.sort()
            self.assertEqual(actual_known_names, expected_known_names)
    
    
    def test_persist_with_scope(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        # need this for `ItemGeneralOne` (related field)
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            
        scope = Scope(
            {
                ItemGeneralOne: {
                    'creators': ['f_integer', 'f_float'],
                    'getters': ['f_integer', 'f_float'],
                }
            },
            collection_id='test_persist_with_scope')
        
        self.assertNotIsInstance(scope[ItemGeneralOne], ItemGeneralOne)
        
        persister = self.persister
        
        # not scoped
        item_one = ItemGeneralOne(f_integer='100', f_float='200')
        item_one_dump = persister.dumps(item_one)
        item_one_load = persister.loads(item_one_dump)
        self.assertIsInstance(item_one_load, ItemGeneralOne)
        
        
        item_one_scoped = scope[ItemGeneralOne](f_integer='300', f_float='400')
        item_one_scoped_dump = persister.dumps(item_one_scoped)
        item_one_scoped_load = persister.loads(item_one_scoped_dump)
        self.assertIsInstance(item_one_scoped_load, scope[ItemGeneralOne])
    
    
    def test_one_to_many_creator(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['two_1_x']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
        
        persister = self.persister
        
        # failed creation    
        item_1 = ItemGeneralOne()
        item_1['two_1_x']  # access creates
        
        items, model_lists = persister.persist(item_1)
        self.assertEqual(len(items), 0)
        self.assertEqual(len(model_lists), 0)
        
        # successful creation
        item_1['two_1_x'].gen(f_integer=10)
        items, model_lists = persister.persist(item_1)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 1)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 1)
        
        model = model_lists[0][0]
        two_1_x = self.get_related_x_to_many(model, 'two_1_x')
        self.assertEqual(len(two_1_x), 1)
        self.assertEqual(two_1_x[0].f_integer, 10)
    
    
    def test_many_to_many_creator(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['two_x_x']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
        
        persister = self.persister
        
        # failed creation    
        item_1 = ItemGeneralOne()
        item_1['two_x_x']  # access creates
        
        items, model_lists = persister.persist(item_1)
        self.assertEqual(len(items), 0)
        self.assertEqual(len(model_lists), 0)
        
        # successful creation
        item_1['two_x_x'].gen(f_integer=10)
        items, model_lists = persister.persist(item_1)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 1)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 1)
        
        model = model_lists[0][0]
        two_x_x = self.get_related_x_to_many(model, 'two_x_x')
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(two_x_x[0].f_integer, 10)
    
    
    def test_x_to_many_getters(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['f_integer']
            getters = ['f_string']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['one_x_x']
        
        persister = self.persister
        
        item_one = ItemGeneralOne(f_integer='1', f_string='str-1')
        item_two = ItemGeneralTwo(f_integer='2', one_x_x=[item_one])

        persister.persist(item_two)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 1)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 1)
        
        item_one = ItemGeneralOne(f_string='str-1')
        item_two = ItemGeneralTwo(f_string='str-2', one_x_x=[item_one])
        
        persister.persist(item_two)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 1)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 1)
        
        model_one = self.get_all_models(self.ModelGeneralOne)[0]
        model_two = self.get_all_models(self.ModelGeneralTwo)[0]
        
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.f_string, 'str-1')
        related_two = self.get_related_x_to_many(model_one, 'two_x_x')
        self.assertEqual(len(related_two), 1)
        self.assertEqual(model_two, related_two[0])
        
        self.assertEqual(model_two.f_integer, 2)
        self.assertEqual(model_two.f_string, 'str-2')
        related_one = self.get_related_x_to_many(model_two, 'one_x_x')
        self.assertEqual(len(related_one), 1)
        self.assertEqual(model_one, related_one[0])
    
    
    def test_null_getters_and_creators(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer', 'f_float', 'f_string', 'f_text'}]
            getters = [{'f_integer', 'f_string'}]
            nullables = ['f_string']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        persister = self.persister
        get_key = lambda model: (model.f_string if model.f_string else '-',
                                 model.f_float,
                                 model.f_integer,
                                 model.f_text)
        
        bulk_item = ItemGeneralOne.Bulk()
        for i in range(3):
            item = bulk_item.gen(
                f_integer=i+1,
                f_string='str-{}'.format(i+1) if i > 1 else None,
                f_text='text-{}'.format(i+1))
            if i != 0:
                item['f_float'] = i+1 + (i+1)/10
        persister.persist(bulk_item)
        
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=get_key)
        model_one_keys = [get_key(model) for model in model_ones]
        expect = [
            # first item cannot be created (missing `f_float`)
            # '-' stands for `None` string (otherwise not sortable)
            ('-', 2.2, 2, 'text-2'),
            ('str-3', 3.3, 3, 'text-3'),
        ]
        self.assertEqual(model_one_keys, expect)
        
        bulk_item = ItemGeneralOne.Bulk()
        for i in range(3):
            item = bulk_item.gen(
                f_integer=i+1,
                f_string='str-{}'.format(i+1) if i > 1 else None,
                f_text='text-{}'.format(i+1 + 100))
            if i != 0:
                item['f_float'] = i+1 + (i+1)/10
        persister.persist(bulk_item)
        
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=get_key)
        model_one_keys = [get_key(model) for model in model_ones]
        expect = [
            ('-', 2.2, 2, 'text-102'),
            ('str-3', 3.3, 3, 'text-103'),
        ]
        self.assertEqual(model_one_keys, expect)
    
    
    def test_one_to_many_getter(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['f_integer']
            getters = ['f_integer', 'two_1_x']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_integer']
        
        persister = self.persister
        
        # creating dummy records
        bulk_item_one = ItemGeneralOne.Bulk()
        bulk_item_two = ItemGeneralTwo.Bulk()
        for f_integer in range(1, 4):  # 1..3
            bulk_item_one.gen(f_integer=1000+f_integer)
            bulk_item_two.gen(f_integer=1000+f_integer)
        
        persister.persist(bulk_item_one)
        persister.persist(bulk_item_two)
        
        #--- one-to-many -------------------------------------------------------
        
        # creating item one
        item_one = ItemGeneralOne(f_integer=2001)  # new
        item_one['two_1_x'].gen(f_integer=1003)  # old
        item_one['two_1_x'].gen(f_integer=1004)  # new
        
        _, model_lists = persister.persist(item_one)
        
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        
        model = model_lists[0][0]
        two_1_x = self.get_related_x_to_many(model, 'two_1_x')
        self.assertEqual(len(two_1_x), 2)
        two_1_x.sort(key=lambda m: m.f_integer)
        self.assertEqual(two_1_x[0].f_integer, 1003)
        self.assertEqual(two_1_x[1].f_integer, 1004)
        
        # updating item one
        item_one = ItemGeneralOne()  # old, will be found by "two_1_x"
        item_one['two_1_x'].gen(f_integer=1004)  # old
        item_one['two_1_x'].gen(f_integer=1005)  # new
        
        _, model_lists = persister.persist(item_one)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 5)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        
        model = model_lists[0][0]
        self.assertEqual(model.f_integer, 2001)
        two_1_x = self.get_related_x_to_many(model, 'two_1_x')
        self.assertEqual(len(two_1_x), 3)
        two_1_x.sort(key=lambda m: m.f_integer)
        self.assertEqual(two_1_x[0].f_integer, 1003)
        self.assertEqual(two_1_x[1].f_integer, 1004)
        self.assertEqual(two_1_x[2].f_integer, 1005)
        
        # empty bulk must be ignored
        item_one = ItemGeneralOne()
        item_one['two_1_x']  # accessing the bulk creates it
        items, model_lists = persister.persist(item_one)
        self.assertEqual(len(items), 0)
        self.assertEqual(len(model_lists), 0)
        # nothing changed
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 5)
        
        
    def test_many_to_many_getter(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['f_integer']
            getters = ['f_integer']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_integer', 'one_x_x']
        
        persister = self.persister
        
        # creating dummy records
        bulk_item_one = ItemGeneralOne.Bulk()
        bulk_item_two = ItemGeneralTwo.Bulk()
        for f_integer in range(1, 4):  # 1..3
            bulk_item_one.gen(f_integer=1000+f_integer)
            bulk_item_two.gen(f_integer=1000+f_integer)
        
        persister.persist(bulk_item_one)
        persister.persist(bulk_item_two)
        
        #--- many-to-many ------------------------------------------------------
        
        # creating item two
        item_two = ItemGeneralTwo(f_integer=2001)  # new
        item_two['one_x_x'].gen(f_integer=1003)  # old
        item_two['one_x_x'].gen(f_integer=1004)  # new
        
        _, model_lists = persister.persist(item_two)
        
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        
        model = model_lists[0][0]
        one_x_x = self.get_related_x_to_many(model, 'one_x_x')
        self.assertEqual(len(one_x_x), 2)
        one_x_x.sort(key=lambda m: m.f_integer)
        self.assertEqual(one_x_x[0].f_integer, 1003)
        self.assertEqual(one_x_x[1].f_integer, 1004)
        
        # updating item one
        item_two = ItemGeneralTwo()  # old, will be found by "one_x_x"
        item_two['one_x_x'].gen(f_integer=1004)  # old
        item_two['one_x_x'].gen(f_integer=1005)  # new
        
        _, model_lists = persister.persist(item_two)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 5)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        
        model = model_lists[0][0]
        self.assertEqual(model.f_integer, 2001)
        one_x_x = self.get_related_x_to_many(model, 'one_x_x')
        self.assertEqual(len(one_x_x), 3)
        one_x_x.sort(key=lambda m: m.f_integer)
        self.assertEqual(one_x_x[0].f_integer, 1003)
        self.assertEqual(one_x_x[1].f_integer, 1004)
        self.assertEqual(one_x_x[2].f_integer, 1005)
        
        # empty bulk must be ignored
        item_two = ItemGeneralTwo()
        item_two['one_x_x']  # accessing the bulk creates it
        items, model_lists = persister.persist(item_two)
        self.assertEqual(len(items), 0)
        self.assertEqual(len(model_lists), 0)
        # nothing changed
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 5)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
    
    
    def test_one_to_many_getter_mix(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['f_integer']
            getters = ['f_integer']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_integer', {'f_float', 'one_1_x'}]
        
        persister = self.persister
        
        # creating dummy records
        bulk_item_one = ItemGeneralOne.Bulk()
        bulk_item_two = ItemGeneralTwo.Bulk()
        for f_integer in range(1, 4):  # 1..3
            bulk_item_one.gen(f_integer=1000+f_integer,
                              f_float=1000.1+f_integer)
            bulk_item_two.gen(f_integer=1000+f_integer,
                              f_float=1000.1+f_integer)
        
        persister.persist(bulk_item_one)
        persister.persist(bulk_item_two)
        
        #--- mix with regular filter -------------------------------------------
        
        # creating item_two
        item_two = ItemGeneralTwo(f_integer=2001, f_float=2001.2)  # new
        item_two['one_1_x'].gen(f_integer=1003)  # old
        item_two['one_1_x'].gen(f_integer=1004)  # new
        
        _, model_lists = persister.persist(item_two)
        
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        
        model = model_lists[0][0]
        one_1_x = self.get_related_x_to_many(model, 'one_1_x')
        self.assertEqual(len(one_1_x), 2)
        one_1_x.sort(key=lambda m: m.f_integer)
        self.assertEqual(one_1_x[0].f_integer, 1003)
        self.assertEqual(one_1_x[1].f_integer, 1004)
        
        # updating item
        # (can find by `f_float`, but item from `one_1_x` is not there)
        item_two = ItemGeneralTwo(f_float=2001.2)
        item_two['one_1_x'].gen(f_integer=1001)  # old
        
        _, model_lists = persister.persist(item_two)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        self.assertEqual(len(model_lists), 0)
        
        # (can find by `one_1_x`, but `f_float` does not match)
        item_two = ItemGeneralTwo(f_float=2002.2)
        item_two['one_1_x'].gen(f_integer=1004)  # old
        
        _, model_lists = persister.persist(item_two)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        self.assertEqual(len(model_lists), 0)
        
        # (empty bulk)
        item_two = ItemGeneralTwo(f_float=2002.2)
        item_two['one_1_x']  # accessing bulk creates it
        
        _, model_lists = persister.persist(item_two)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        
        # now update must succeed
        item_two = ItemGeneralTwo(f_float=2001.2)
        item_two['one_1_x'].gen(f_integer=1004)  # old
        item_two['one_1_x'].gen(f_integer=1005)  # new
        
        _, model_lists = persister.persist(item_two)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 5)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        model = model_lists[0][0]
        one_1_x = self.get_related_x_to_many(model, 'one_1_x')
        self.assertEqual(len(one_1_x), 3)
        one_1_x.sort(key=lambda m: m.f_integer)
        self.assertEqual(one_1_x[0].f_integer, 1003)
        self.assertEqual(one_1_x[1].f_integer, 1004)
        self.assertEqual(one_1_x[2].f_integer, 1005)
        
    
    def test_many_to_many_getter_mix(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['f_integer']
            getters = ['f_integer']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_integer', {'f_float', 'one_x_x'}]
        
        persister = self.persister
        
        # creating dummy records
        bulk_item_one = ItemGeneralOne.Bulk()
        bulk_item_two = ItemGeneralTwo.Bulk()
        for f_integer in range(1, 4):  # 1..3
            bulk_item_one.gen(f_integer=1000+f_integer,
                              f_float=1000.1+f_integer)
            bulk_item_two.gen(f_integer=1000+f_integer,
                              f_float=1000.1+f_integer)
        
        persister.persist(bulk_item_one)
        persister.persist(bulk_item_two)
        
        #--- mix with regular filter -------------------------------------------
        
        # creating item_two
        item_two = ItemGeneralTwo(f_integer=2001, f_float=2001.2)  # new
        item_two['one_x_x'].gen(f_integer=1003)  # old
        item_two['one_x_x'].gen(f_integer=1004)  # new
        
        _, model_lists = persister.persist(item_two)
        
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        
        model = model_lists[0][0]
        one_x_x = self.get_related_x_to_many(model, 'one_x_x')
        self.assertEqual(len(one_x_x), 2)
        one_x_x.sort(key=lambda m: m.f_integer)
        self.assertEqual(one_x_x[0].f_integer, 1003)
        self.assertEqual(one_x_x[1].f_integer, 1004)
        
        # updating item
        # (can find by `f_float`, but item from `one_x_x` is not there)
        item_two = ItemGeneralTwo(f_float=2001.2)
        item_two['one_x_x'].gen(f_integer=1001)  # old
        
        _, model_lists = persister.persist(item_two)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        self.assertEqual(len(model_lists), 0)
        
        # (can find by `one_x_x`, but `f_float` does not match)
        item_two = ItemGeneralTwo(f_float=2002.2)
        item_two['one_x_x'].gen(f_integer=1004)  # old
        
        _, model_lists = persister.persist(item_two)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        self.assertEqual(len(model_lists), 0)
        
        # (empty bulk)
        item_two = ItemGeneralTwo(f_float=2002.2)
        item_two['one_x_x']  # accessing bulk creates it
        
        _, model_lists = persister.persist(item_two)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 4)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        
        # now update must succeed
        item_two = ItemGeneralTwo(f_float=2001.2)
        item_two['one_x_x'].gen(f_integer=1004)  # old
        item_two['one_x_x'].gen(f_integer=1005)  # new
        
        _, model_lists = persister.persist(item_two)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralOne)), 5)
        self.assertEqual(len(self.get_all_models(self.ModelGeneralTwo)), 4)
        
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        model = model_lists[0][0]
        one_x_x = self.get_related_x_to_many(model, 'one_x_x')
        self.assertEqual(len(one_x_x), 3)
        one_x_x.sort(key=lambda m: m.f_integer)
        self.assertEqual(one_x_x[0].f_integer, 1003)
        self.assertEqual(one_x_x[1].f_integer, 1004)
        self.assertEqual(one_x_x[2].f_integer, 1005)
    
    
    def test_related_overwrite_with_none(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['two_x_1']
            getters = ['two_x_1']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_integer']
        
        item_one = ItemGeneralOne(two_x_1=ItemGeneralTwo(f_integer=1))
        _, model_list = self.persister.persist(item_one)
        model_one = model_list[0][0]
        self.assertEqual(model_one.two_x_1.f_integer, 1)
        
        item_one = ItemGeneralOne(two_x_1=None)
        _, model_list = self.persister.persist(item_one)
        self.assertEqual(len(model_list), 1)
        model_one = model_list[0][0]
        self.assertIsNone(model_one.two_x_1)
    
    
    def test_related_none_creator_and_getter(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['two_1_1']
            getters = ['two_1_1']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_integer']
        
        # no value - no model created
        item_one = ItemGeneralOne()
        _, model_list = self.persister.persist(item_one)
        self.assertEqual(len(model_list), 0)
        
        # can use `None` as creator
        item_one = ItemGeneralOne(two_1_1=None)
        _, model_list = self.persister.persist(item_one)
        self.assertEqual(len(model_list), 1)
        model_one = model_list[0][0]
        self.assertTrue(not hasattr(model_one, 'two_1_1') or
                        model_one.two_1_1 is None)
        
        model_one_list = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(model_one_list), 1)
        
        # updating using `None` getter
        original_model = model_one
        item_one = ItemGeneralOne(two_1_1=None, f_integer=10)
        _, model_list = self.persister.persist(item_one)
        self.assertEqual(len(model_list), 1)
        model_one = model_list[0][0]
        self.assertEqual(original_model, model_one)
        self.assertTrue(not hasattr(model_one, 'two_1_1') or
                        model_one.two_1_1 is None)
        self.assertEqual(model_one.f_integer, 10)
        
        model_one_list = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(model_one_list), 1)


from save_to_db.adapters.utils.adapter_manager import get_adapter_cls
from save_to_db.adapters.utils.column_type import ColumnType
from save_to_db.adapters.utils.relation_type import RelationType

from save_to_db.utils.test_base import TestBase



class TestFieldsAndRelations(TestBase):
    """ Tests that adapter correctly identifies model column types and
    relationship types of foreign keys.
    """
    
    # models that are needed for tests to run
    ModelFieldTypes = None
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    
    def test_iter_fields(self):
        #--- test column types ---
        model_cls = self.ModelFieldTypes
        adapter_cls = get_adapter_cls(model_cls)
        
        required_column_types = [ct for ct in ColumnType.__members__.values()
                                    if ct != ColumnType.OTHER]

        for col_name, col_type in adapter_cls.iter_fields(model_cls):
            if col_name == 'id':
                continue  # ignoring "id" field
            if col_type in required_column_types:
                required_column_types.remove(col_type)
            
            col_type_from_name = col_name.split('_', 2)[0].upper()
            self.assertEqual(col_type_from_name, col_type.name,
                             'Wrong column type for: {}'.format(col_name))

        self.assertTrue(len(required_column_types) == 0,
                        'Not all columns types were defined, missing: {}'
                            .format(', '.join(t.name for t in
                                              required_column_types)))
        
        #--- foreign keys must not be reported ---
        model_cls = self.ModelGeneralOne
        expected_fields = {
            ('id', ColumnType.INTEGER),
            ('f_binary', ColumnType.BINARY),
            ('f_boolean', ColumnType.BOOLEAN),
            ('f_string', ColumnType.STRING),
            ('f_text', ColumnType.TEXT),
            ('f_integer', ColumnType.INTEGER),
            ('f_float', ColumnType.FLOAT),
            ('f_decimal', ColumnType.DECIMAL),
            ('f_date', ColumnType.DATE),
            ('f_time', ColumnType.TIME),
            ('f_datetime', ColumnType.DATETIME),
        }
        
        actual_fields = set(
            (fname, ftype)
            for fname, ftype, in adapter_cls.iter_fields(model_cls))
        
        self.assertFalse(expected_fields - actual_fields,
                         'Not all columns were reported')
        self.assertFalse(actual_fields - expected_fields,
                         'Extra columns were reported')
        
    
    def test_iter_relations(self):
        model_one = self.ModelGeneralOne
        model_two = self.ModelGeneralTwo
        
        adapter_cls = get_adapter_cls(model_one)
        
        #--- model_one ---
        
        expected_relation_set = {
            # self relations
            ('parent_1_1', model_one, RelationType.ONE_TO_ONE, 'child_1_1'),
            ('child_1_1', model_one, RelationType.ONE_TO_ONE, 'parent_1_1'),
            ('parent_x_1', model_one, RelationType.MANY_TO_ONE, 'child_1_x'),
            ('child_1_x', model_one, RelationType.ONE_TO_MANY, 'parent_x_1'),
            ('parent_x_x', model_one, RelationType.MANY_TO_MANY, 'child_x_x'),
            ('child_x_x', model_one, RelationType.MANY_TO_MANY, 'parent_x_x'),
            # relations with other
            ('two_1_1', model_two, RelationType.ONE_TO_ONE, 'one_1_1'),
            ('two_1_x', model_two, RelationType.ONE_TO_MANY, 'one_x_1'),
            ('two_x_1', model_two, RelationType.MANY_TO_ONE, 'one_1_x'),
            ('two_x_x', model_two, RelationType.MANY_TO_MANY, 'one_x_x'),
        }
        actual_relation_set = set(adapter_cls.iter_relations(model_one))

        # not all relations reported?
        not_reported_relations = expected_relation_set - actual_relation_set
        self.assertFalse(
            bool(not_reported_relations),
            'Not all relations reported for {}, not reported: {}'.format(
                 model_one.__name__, str(not_reported_relations)))
        
        # reported extra relations?
        extra_relations = actual_relation_set - expected_relation_set
        self.assertFalse(extra_relations,
                         'Extra relations reported for {}'.format(
                             model_one.__name__))
        
        # relation reported twice?
        actual_relation_list = list(adapter_cls.iter_relations(model_one))
        actual_relation_list.sort()
        for actual_relation in actual_relation_set:
            actual_relation_list.remove(actual_relation)
        actual_relation_list = list(set(actual_relation_list))
        actual_relation_list.sort()
        self.assertFalse(actual_relation_list,
                         'Relations reported more then once for {}'.format(
                             model_one.__name__))
        
        
        #--- model_two ---

        expected_relation_set = {
            # self relations
            ('parent_1_1', model_two, RelationType.ONE_TO_ONE, 'child_1_1'),
            ('child_1_1', model_two, RelationType.ONE_TO_ONE, 'parent_1_1'),
            ('parent_x_1', model_two, RelationType.MANY_TO_ONE, 'child_1_x'),
            ('child_1_x', model_two, RelationType.ONE_TO_MANY, 'parent_x_1'),
            ('parent_x_x', model_two, RelationType.MANY_TO_MANY, 'child_x_x'),
            ('child_x_x', model_two, RelationType.MANY_TO_MANY, 'parent_x_x'),
            # relations with other
            ('one_1_1', model_one, RelationType.ONE_TO_ONE, 'two_1_1'),
            ('one_1_x', model_one, RelationType.ONE_TO_MANY, 'two_x_1'),
            ('one_x_1', model_one, RelationType.MANY_TO_ONE, 'two_1_x'),
            ('one_x_x', model_one, RelationType.MANY_TO_MANY, 'two_x_x'),
        }
        actual_relation_set = set(adapter_cls.iter_relations(model_two))
        
        # not all relations reported?
        not_reported_relations = expected_relation_set - actual_relation_set
        self.assertFalse(not_reported_relations,
                         'Not all relations reported for {}'.format(
                             model_two.__name__))
        
        # reported extra relations?
        extra_relations = actual_relation_set - expected_relation_set
        self.assertFalse(extra_relations,
                         'Extra relations reported for {}'.format(
                             model_two.__name__))
        
        # relation reported twice?
        actual_relation_list = list(adapter_cls.iter_relations(model_two))
        actual_relation_list.sort()
        for actual_relation in actual_relation_set:
            actual_relation_list.remove(actual_relation)
        actual_relation_list = list(set(actual_relation_list))
        actual_relation_list.sort()
        self.assertFalse(actual_relation_list,
                         'Relations reported more then once for {}'.format(
                             model_two.__name__))
    
    
    def test_table_fullname(self):
        adapter_cls = get_adapter_cls(self.ModelFieldTypes)
        
        table_name = adapter_cls.get_table_fullname(self.ModelFieldTypes)
        self.assertEqual(table_name, 'model_field_types')
        
        table_name = adapter_cls.get_table_fullname(self.ModelGeneralOne)
        self.assertEqual(table_name, 'model_general_one')
        
        table_name = adapter_cls.get_table_fullname(self.ModelGeneralTwo)
        self.assertEqual(table_name, 'model_general_two')
    
    
    
    
    
    
    


from save_to_db.core.item import Item
from save_to_db.exceptions import MultipleModelsMatch, MultipleItemsMatch
from save_to_db.utils.test_base import TestBase



class TestExceptions(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    ModelManyRefsOne = None
    ModelManyRefsTwo = None
    
    ModelAutoReverseOne = None
    ModelAutoReverseTwoA = None
    ModelAutoReverseTwoB = None
    ModelAutoReverseThreeA = None
    ModelAutoReverseThreeB = None
    ModelAutoReverseFourA = None
    ModelAutoReverseFourB = None
    
    
    def test_multiple_items_match_exception(self):
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        persister = self.persister
        
        # check for existing models
        persister.persist(ItemGeneralOne(f_integer=10, f_float=1))
        persister.persist(ItemGeneralOne(f_integer=20, f_float=2))
        
        bulk = ItemGeneralOne.Bulk()
        # two items will get the same model
        bulk.gen(f_integer=10, f_string='str-value-one')
        bulk.gen(f_float=1, f_string='str-value-two')
        
        with self.assertRaises(MultipleItemsMatch):
            persister.persist(bulk)
            
        # check for newly created models
        bulk = ItemGeneralOne.Bulk()
        bulk.gen(f_integer=10, f_float=1)
        bulk.gen(f_integer=20, f_float=1)
        
        with self.assertRaises(MultipleItemsMatch):
            persister.persist(bulk)
        
        # trying to create the same model twice
        bulk = ItemGeneralOne.Bulk()
        bulk.gen(f_integer=100)
        bulk.gen(f_integer=100)
        with self.assertRaises(MultipleItemsMatch):
            persister.persist(bulk)
    
    
    def test_multiple_models_match_exception(self):
        model_one_cls = self.ModelGeneralOne
        model_two_cls = self.ModelGeneralTwo
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float', 'f_text'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float', 'one_x_1'}]
        
        persister = self.persister
        
        # with simple fields ---------------------------------------------------
        # creating two items with different integer fields but same other fields
        item = ItemGeneralOne(f_integer='100', f_float='200.200')
        persister.persist(item)
        item = ItemGeneralOne(f_integer='200', f_float='200.200')
        persister.persist(item)
        
        # updating to get 'f_float' and 'f_text' have the same values for
        # different models 
        item = ItemGeneralOne(f_integer='100', f_text='text-1')
        persister.persist(item)
        item = ItemGeneralOne(f_integer='200', f_text='text-1')
        persister.persist(item)
        
        self.assertEqual(len(self.get_all_models(model_one_cls)), 2)
        
        # this must get two models from database
        item = ItemGeneralOne(f_integer='100',
                              f_float='200.200', f_text='text-1')
        with self.assertRaises(MultipleModelsMatch):
            persister.persist(item)
        
        # still 2 models in database
        self.assertEqual(len(self.get_all_models(model_one_cls)), 2)
        
        # using relations ------------------------------------------------------
        item = ItemGeneralTwo(f_integer='100', f_float='200.200')
        persister.persist(item)
        item = ItemGeneralTwo(f_integer='200', f_float='200.200')
        persister.persist(item)
        
        item_one = ItemGeneralOne(f_integer='300')  # new item one
        item = ItemGeneralTwo(f_integer='100', one_x_1=item_one)
        persister.persist(item)
        item = ItemGeneralTwo(f_integer='200', one_x_1=item_one)
        persister.persist(item)
        
        self.assertEqual(len(self.get_all_models(model_one_cls)), 3)
        self.assertEqual(len(self.get_all_models(model_two_cls)), 2)

        # this must get two models from database
        item = ItemGeneralTwo(f_float='200.200', one_x_1=item_one)
        with self.assertRaises(MultipleModelsMatch):
            persister.persist(item)        
            
            
    def test_multiple_models_match_x_to_1_exception(self):
        model_one_cls = self.ModelGeneralOne
        
        class ItemGeneralOne(Item):
            allow_multi_update = True
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float', 'f_text'}]
        
        class ItemGeneralTwo(Item):
            allow_multi_update = True
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float', 'one_x_1'}]
        
        persister = self.persister
        
        # creating two items with different integer fields but same other fields
        item = ItemGeneralOne(f_integer='100', f_float='200.200',
                              f_boolean=True)
        persister.persist(item)
        item = ItemGeneralOne(f_integer='200', f_float='200.200',
                              f_boolean=False)
        persister.persist(item)
        
        # updating to get 'f_float' and 'f_text' have the same values for
        # different models 
        item_one = ItemGeneralOne(f_integer='100', f_text='text-1')
        persister.persist(item_one)
        item_one = ItemGeneralOne(f_integer='200', f_text='text-1')
        persister.persist(item_one)
        
        self.assertEqual(len(self.get_all_models(model_one_cls)), 2)
        
        #--- set two models to x-to-one field must fail ------------------------
        item_two = ItemGeneralTwo(f_integer='200')
        item_two['one_1_1'](f_float='200.200', f_text='text-1')
        with self.assertRaises(MultipleModelsMatch):
            persister.persist(item_two)


""" This module contains exceptions for
:py:class:`~save_to_db.adapters.adapter_base.AdapterBase` class.
"""

class AdapterException(Exception):
    """ General exception for
    :py:class:`~save_to_db.adapters.adapter_base.AdapterBase` class.
    """


class CannotMergeModels(AdapterException):
    """ Raised when merging two or more ORM models into one is impossible. """


class MergeModelsNotAllowed(AdapterException):
    """ Raised when merging ORM models is forbidden. """


from enum import Enum


class ColumnType(Enum):
    """ Enumeration class for all possible DB types with which this library can
    deal.
    
    .. note:: See the source code for helper methods that each enumeration has.
    """
    
    # for Python versions prior to 3.6
    def __new__(cls):
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj
    
    # binary
    BINARY = ()         #: Any type of binary data
    
    # boolean
    BOOLEAN = ()        #: Boolean value
    
    # strings
    STRING = ()         #: String value, like *char* or *varchar*
    TEXT = ()           #: String unlimited in length
    
    # numbers
    INTEGER = ()        #: Any type of integer data
    FLOAT = ()          #: Any type of numbers with fraction (float)
    DECIMAL = ()        #: Any type of numbers with fraction (decimal.Decimal)
    
    # dates and times
    DATE = ()           #: Date value
    TIME = ()           #: Time value
    DATETIME = ()       #: Date and time value (*timestamp*)
    
    # unknown
    OTHER = ()          #: Unknown data type
    
    
    def is_num(self):
        return self in (ColumnType.INTEGER,
                        ColumnType.FLOAT,
                        ColumnType.DECIMAL)
    
    def is_str(self):
        return self in (ColumnType.STRING, ColumnType.TEXT)
    
    



""" This module contains exceptions for
:py:class:`~save_to_db.core.item_base.ItemBase` class related to persistence.
"""


class ItemPersistError(Exception):
    """ General exception for :py:class:`~save_to_db.core.item_base.ItemBase`
    persistence.
    """


class MultipleItemsMatch(ItemPersistError):
    """ Raised when multiple items match the same model. """
    
    
class MultipleModelsMatch(ItemPersistError):
    """ Raised when multiple models match the same item:
        
        - When `allow_multi_update` set to `False` for the item.
        - When trying to set multiple items to x-to-one relationship.
    """


class CannotMergeModelsForItem(ItemPersistError):
    """ Raised when merging two or more ORM models into one is impossible. """


class CannotClearRequiredFieldInRelation(ItemPersistError):
    """ Raise when trying to clear one-to-many field using "replace_x_to_many"
    setting of item, but field is required on the other side.
    """


from save_to_db.utils.test_base import TestBase
from .config import Base, engine, session


class SqlalchemyTestBase(TestBase):
    
    def setUp(self):
        # preparing database
        session.rollback()
        Base.metadata.drop_all(engine)
        Base.metadata.create_all(engine)
        
        super().setUp()
        



import copy
from unittest import TestCase
from save_to_db.core import signals
from save_to_db.core.item_cls_manager import item_cls_manager
from save_to_db.core.scope import Scope




class TestBase(TestCase):
    
    item_cls_manager = item_cls_manager
    
    @classmethod
    def setUpClass(cls):
        cls.item_cls_manager.clear()
        Scope.delete_all_scopes()
        super().setUpClass()
        
        
    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.item_cls_manager.clear()
        Scope.delete_all_scopes()
    
    
    def setUp(self):
        self.item_cls_manager.autogenerate = False
        self.__item_cls_manager_collection_before = \
            set(self.__class__.item_cls_manager.collection)
        self.__collection_classes_before = copy.copy(Scope.collection_classes)
        super().setUp()
        
    
    def tearDown(self):
        super().tearDown()
        self.__class__.item_cls_manager.collection = \
            self.__item_cls_manager_collection_before
        Scope.collection_classes.clear()
        Scope.collection_classes.update(self.__collection_classes_before)
        for signal in signals.all_signals:
            signal.clear()

    
    def get_all_models(self, model_cls, sort_key=None):
        db_adapter = self.persister.db_adapter
        return db_adapter.get_all_models(model_cls, sort_key=sort_key)
        
        
    def get_related_x_to_many(self, model, field_name, sort_key=None):
        db_adapter = self.persister.db_adapter
        result = list(db_adapter.get_related_x_to_many(model, field_name))
        if sort_key:
            result.sort(key=sort_key)
        return result



from itertools import chain
from save_to_db.core.item_base import ItemBase
from save_to_db.exceptions import ItemsNotTheSame


def merge_items(top_item, cls_items, cls_default_items, all_items):
    """ Merges items that pull the same model from the database into a single
    items.
    
    :param top_item: the item that is being processed.
    :param items: List of items in which items referring to the same model
        must be merged.
    :param all_items: all single items in structure.
    """

    all_cls_items = cls_items[:]
    all_cls_items.extend(cls_default_items)
    
    merged_items = []
    item_to_merged = []
    while True:
        for item in all_cls_items:
            item_merged = False
            
            if item in merged_items or not item.getters:
                continue
            
            for group in item.getters:
                
                use_group = True
                for field_name in group:
                    if field_name not in item.data:
                        use_group = False
                        break  
                if not use_group:
                    continue
                
                for next_item in all_cls_items:
                    if next_item is item or next_item in merged_items:
                        continue
                    
                    do_merge = True
                    for key in group:
                        if key not in next_item or \
                                (not isinstance(item[key], ItemBase) and
                                 item[key] != next_item[key]) or \
                                (isinstance(item[key], ItemBase) and
                                 not __same_items(item[key], next_item[key])):
                            do_merge = False
                            break

                    if not do_merge:
                        continue
                    
                    __merge_item_into_item(item, next_item)
                    merged_items.append(next_item)
                    item_to_merged.append([item, next_item])
                    item_merged = True
            
            if item_merged:
                break
        
        # replacing merged_items with the extended item in single items
        for _, merged_item in item_to_merged:
            for one_item in all_items:
                for key, value in one_item.data.items():
                    if value is merged_item:
                        one_item[key] = item
            
            all_cls_items.remove(merged_item)
            if merged_item in cls_items:
                cls_items.remove(merged_item)
        
        # replacing in bulk items
        __replace_in_bulk_items(top_item, item_to_merged, processed_items=[])
            
        merged_items.clear()
        item_to_merged.clear()
        
        if not item_merged:
            break


def __replace_in_bulk_items(item, item_to_merged, processed_items):
    if item in processed_items:
        return
    processed_items.append(item)
    
    if item.is_single_item():
        for f_key, relation in item.relations.items():
            if f_key not in item or \
                    not relation['relation_type'].is_x_to_many():
                continue
            __replace_in_bulk_items(item[f_key], item_to_merged,
                                    processed_items=processed_items)
    else:
        for extened_item, merged_item in item_to_merged:
            if merged_item in item:
                index = item.bulk.index(merged_item)
                item.remove(merged_item)
                item.add_at_index(index, extened_item)
            
            # updating defaults
            for key, default_value in item.data.items():
                if not isinstance(default_value, ItemBase):
                    continue
                 
                if default_value is merged_item:
                    item[key] = extened_item
                    
                if default_value.is_bulk_item():
                    __replace_in_bulk_items(default_value, item_to_merged,
                                            processed_items=processed_items)
        
        for item_in_bulk in item:
            __replace_in_bulk_items(item_in_bulk, item_to_merged,
                                    processed_items=processed_items)


def __same_items(item_one, item_two, _no_check_pairs=None):
    if item_one.is_bulk_item():
        return True
        
    _no_check_pairs = _no_check_pairs or []
    
    for key in set(chain(item_one.data.keys(), item_two.data.keys())):
        if key not in item_one.data or key not in item_two.data:
            continue
        
        if key in item_one.relations and \
                item_one.relations[key]['relation_type'].is_x_to_many():
            continue
        
        if key not in item_one.data or key not in item_two.data or \
                item_two.data[key] != item_one.data[key]:
            
            if key in item_one.relations:
                pair = [item_one, item_two]
                if pair in _no_check_pairs:
                    continue
                _no_check_pairs.append(pair)
                if type(item_one.data[key]).allow_merge_items:
                    if __same_items(item_one.data[key],
                                    item_two.data[key],
                                    _no_check_pairs=_no_check_pairs):
                        continue
            return False
        
    return True


def __merge_item_into_item(extended_item, merged_item):
    if not __same_items(extended_item, merged_item):
        raise ItemsNotTheSame(extended_item, merged_item)
    
    for key in merged_item.data:
        is_x_to_many = False
        if key in extended_item.relations:
            is_x_to_many = \
                extended_item.relations[key]['relation_type'].is_x_to_many()
        
        if not is_x_to_many:
            if key not in extended_item.data:
                extended_item[key] = merged_item[key]
        else:
            for item in merged_item[key].bulk:
                extended_item[key].add(item)
                
                


from .item import ItemBase
from .item_contructor import complete_item_structure

from .utils.mapper import mapper

from save_to_db.exceptions import BulkItemOneToXDefaultError


class BulkItem(ItemBase):
    """ This class deals with instances of :py:class:`~.item.Item` in chunks.
    It can create or update multiple database rows using single query, e.g. it
    can persist multiple items at once.
    
    .. note::
        You get items values from a bulk item like this:
        
            - `bulk[number]` return item at a given index `number`;
            - `bulk[string]` return default value for a key `string`;
            - `bulk[slice]` (e.g. `bulk[1:2]`) returns python list of items
              containing specified items.
              
    .. note::
        Defaults order is based on the number of '__' the key contains,
        shorter keys are prioritized. This is due to the fact that default
        values can be another items.
    
    :param item_cls: A subclass of :py:class:`~.item.Item` that
        this class deals with.
    :param \*\*kwargs: Values that will be saved as **default** item data.
    """
    
    #--- special methods -------------------------------------------------------
    
    def __init__(self, item_cls, **kwargs):
        self.bulk = []
        self.item_cls = item_cls
        super().__init__()
        
        for key, value in kwargs.items():
            self[key] = value  # this will trigger `__setitem__` function
        
    
    def __setitem__(self, key, value):
        real_key = self.item_cls._get_real_keys(key)
        real_key_str = '__'.join(real_key)
        
        # value one-to-x must not be an item
        if isinstance(value, ItemBase):
            # getting last relation type
            item_cls = self.item_cls
            for key in real_key:
                if key not in item_cls.relations:
                    break
                relation = item_cls.relations[key]
                item_cls = relation['item_cls']
                
                if relation['relation_type'].is_one_to_x():
                    raise BulkItemOneToXDefaultError(self.item_cls, key,
                                                     real_key_str)
        
        self.data[real_key_str] = value
        
    
    def __getitem__(self, key):
        if isinstance(key, int) or isinstance(key, slice):
            return self.bulk[key]
        
        real_keys = self.item_cls._get_real_keys(key)
        return self._get_direct(real_keys)
    
    
    def _get_direct(self, real_keys):
        real_key_str = '__'.join(real_keys)
        if real_key_str in self.data:
            return self.data[real_key_str]
        
        item_cls = self.item_cls
        previous_cls = None
        for real_key in real_keys:
            previous_cls = item_cls
            if real_key in item_cls.relations:
                item_cls = item_cls.relations[real_key]['item_cls']
            else:
                raise KeyError(real_key)
        
        if len(real_keys) == 1:
            self[real_key_str] = self.item_cls._cls_genitem(real_keys[-1])
        else:
            self[real_key_str] = previous_cls._cls_genitem(real_keys[-1])
            
        return self.data[real_key_str]
    
    
    def __delitem__(self, key):
        real_key = self.item_cls._get_real_keys(key, as_string=True)
        del self.data[real_key]
    
    
    def __contains__(self, key):
        if isinstance(key, ItemBase):
            return key in self.bulk
        
        try:
            real_keys = self.item_cls._get_real_keys(key)
        except KeyError:
            return False
        return self._contains_direct(real_keys)
    
    
    def _contains_direct(self, real_keys):
        return '__'.join(real_keys) in self.data
    
    
    def __iter__(self):
        for item in self.bulk:
            yield item
    
    
    def __len__(self):
        return len(self.bulk)
        
    
    #--- utility methods -------------------------------------------------------
    
    def to_dict(self, revert=False):
        return self._to_dict(revert=revert,
                             _item_to_dict={}, _address_to_item={})
        
        
    def _to_dict(self, revert, _item_to_dict, _address_to_item):
        self_address = id(self)
        if self_address in _item_to_dict:
            if self_address not in _address_to_item:
                _address_to_item[self_address] = len(_address_to_item) + 1
            self_id = _address_to_item[self_address]
            _item_to_dict[self_address]['id'] = self_id
            return {
                'id': self_id,
            }
        
        result = {
            'defaults': {},
            'bulk': [],
        }
        _item_to_dict[self_address] = result
        
        # defaults first in `_load_dict()` as well
        defaults = result['defaults']
        for key, value in self.data.items():
            if not isinstance(value, ItemBase):
                if revert:
                    value = self.item_cls.revert_field(key, value,
                                                       aliased=False)
                defaults[key] = value
            else:
                defaults[key] = value._to_dict(revert=revert,
                                               _item_to_dict=_item_to_dict,
                                               _address_to_item=_address_to_item)
        
        bulk = result['bulk']
        for item in self.bulk:
            bulk.append(item._to_dict(revert=revert,
                                      _item_to_dict=_item_to_dict,
                                      _address_to_item=_address_to_item))
        
        return result
    
    
    def load_dict(self, data):
        return self._load_dict(data, _id_to_item={})
    

    def _load_dict(self, data, _id_to_item):
        if 'id' in data:
            if data['id'] not in _id_to_item:
                _id_to_item[data['id']] = self.item_cls.Bulk()
                
            bulk = _id_to_item[data['id']]
            if 'defaults' not in data:
                return _id_to_item[data['id']]
        else:
            bulk = self.item_cls.Bulk()
        
        # defaults first in `_to_dict()` as well
        for key, value in data['defaults'].items():
            # getting relation class
            cur_cls = bulk.item_cls
            for cur_key in key.split('__'):
                if cur_key in cur_cls.relations:
                    cur_cls = cur_cls.relations[cur_key]['item_cls']
                else:
                    cur_cls = None
                    break
            
            if cur_cls:  # relation
                bulk[key] = bulk[key]._load_dict(value, _id_to_item)
            else:
                bulk[key] = value
        
        # bulk
        for dict_wrapper in data['bulk']:
            bulk.add(bulk.item_cls()._load_dict(dict_wrapper, _id_to_item))
        
        return bulk
    
    
    #--- properties ------------------------------------------------------------
    
    @property
    def model_cls(self):
        """ Property that returns `model_cls` attribute of the `item_cls`
        class.
        """
        return self.item_cls.model_cls
    
    def get_item_cls(self):
        return self.item_cls
    
    def is_scoped(self):
        return self.item_cls.metadata['collection_id'] != None
    
    def get_collection_id(self):
        return self.item_cls.metadata['collection_id']
    
    @property
    def relations(self):
        return self.item_cls.relations
    
    
    #--- main methods ----------------------------------------------------------
    
    def add(self, *items):
        """ Adds `items` to the bulk.
        
        :param \*items: List of instances of :py:class:`~.item_base.ItemBase`
            class to be added to the bulk.
        """
        self.add_at_index(None, *items)
    
    
    def add_at_index(self, index, *items):
        """ Adds `items` to the bulk at `index` position.
        If item already in the bulk, nothing will happen.
        
        :param index: Starting position in the bulk list.
        :param \*items: List of instances of :py:class:`~.item_base.ItemBase`
            class to be added to the bulk.
        """
        for index_shift, item in enumerate(items):
            if item not in self.bulk:
                if index is None:
                    self.bulk.append(item)
                else:
                    self.bulk.insert(index+index_shift, item)
                
                mapper.item_added_to_bulk(item, self)
    
    
    def gen(self, *args, **kwargs):
        """ Creates a :py:class:`~.item.Item` instance and adds it to the bulk.
        
        :param \*args: Positional arguments that are passed to the item
            constructor.
        :param \*\*kwargs: Keyword arguments that are passed to the item
            constructor.
        :returns: :py:class:`~.item.Item` instance.
        """
        item = self.item_cls(*args, **kwargs)
        self.add(item)
        
        return item
    
    
    def remove(self, *items):
        """ Removes `item` from the bulk.
        
        :param \*items: List of instances of :py:class:`~.item_base.ItemBase`
            class to be removed from the bulk.
        """
        for item in items:
            if item in self.bulk:
                self.bulk.remove(item)
                mapper.item_removed_from_bulk(item, self)
    
    
    def as_list(self):
        return self.bulk
    
    
    def is_single_item(self):
        return False
    
    
    def is_bulk_item(self):
        return True
    
    
    def revert(self):
        return self._revert(_procesed_items=[])
    
    
    def _revert(self, _procesed_items=[]):
        if self in _procesed_items:
            return
        _procesed_items.append(self)
        
        for key, value in self.data.items():
            if not isinstance(value, ItemBase):
                self.data[key] = self.item_cls.revert_field(key, value,
                                                            aliased=False)
            else:
                self.data[key]._revert(_procesed_items=_procesed_items)
        
        for item in self.bulk:
            item._revert(_procesed_items=_procesed_items)
    
    
    def process(self):
        self._process(_procesed_items=[])
        return complete_item_structure(self)
    

    def _process(self, _procesed_items):
        if self in _procesed_items:
            return
        _procesed_items.append(self)
        
        keys = list(self.data.keys())
        keys.sort(key=lambda key: key.count('__'))
        for key in keys:
            value = self.data[key]
            
            end_item_cls = self.item_cls
            end_relation = None
            for subkey in key.split('__'):
                if subkey in end_item_cls.relations:
                    end_relation = end_item_cls.relations[subkey]
                    end_item_cls = end_relation['item_cls']
                else:
                    end_relation = None
            
            if not end_relation:  # not a relation
                self.data[key] = self.item_cls.process_field(key, value,
                                                             aliased=False)
            else:
                if not isinstance(value, list) or \
                        not end_relation['relation_type'].is_x_to_many():
                    value._process(_procesed_items=_procesed_items)
                    self.data[key] = value
                else:
                    bulk_item = self.item_cls.relations[key]['item_cls'].Bulk()
                    bulk_item.add(*value)
                    bulk_item._process(_procesed_items=_procesed_items)
                    self.data[key] = bulk_item
            
            for item in self.bulk:
                if key in item:
                    continue
                
                item[key] = self.data[key]
        
        for item in self.bulk:
            item._process(_procesed_items=_procesed_items)
        
        


""" This module contains exceptions for
:py:class:`~save_to_db.core.item_base.ItemBase` related to fields usage.
"""


class ItemFieldError(ValueError):
    """ General exception for :py:class:`~save_to_db.core.item_base.ItemBase`
    field usage.
    """

class WrongAlias(ItemFieldError, KeyError):
    """ Raise when wrong alias is used. """


class BulkItemOneToXDefaultError(ItemFieldError):
    """ Raised when trying to assing an one-to-x default relationship value
    to a bulk item.
    """


""" This module contains exceptions that are raise when there is an error
processing an item.
"""

class ItemProcessError(Exception):
    """ General exception for processing instances of
    :py:class:`~save_to_db.core.item_base.ItemBase` class.
    """


class ItemsNotTheSame(ItemProcessError):
    """ Raised when trying to merge items that cannot be merged during
    processing items in a bulk item that can potentially refer to
    the same record in a database.
    """



def select(item, key):
    """ Returns all values under the `key` from `item`. If item contains
    refrences to x-to-many field, the items in that field will be traversed,
    example:
    
        .. code-block:: Python
        
            item_one = ItemOne()
            # 'two_1_x' is a one-to-many relation
            item_two = item['two_1_x'].gen()
            # 'one_x_x' is a many-to-many relation
            item_two['one_x_x'].gen(integer_value=1)
            item_two['one_x_x'].gen(integer_value=2)
            
            result = item_one.select('two_1_x__one_x_x__integer_value')
            print(result)  # outputs: `[1, 2]`

    :param item: An instance of :py:class:`~.item.Item` or
        :py:class:`~.bulk_item.BulkItem`
    :param key: A key under which all values will be collected.
    :returns: List of collected values.
    """
    if not isinstance(key, list):
        if item.is_bulk_item():
            key = item.item_cls._get_real_keys(key)
        else:
            key = item._get_real_keys(key)
    
    result = []
    first_key = key[0]
    if len(key) > 1:
        for one_item in item.as_list():
            if item.is_bulk_item():
                result.extend(select(one_item, key))
            else:
                result.extend(select(one_item[first_key], key[1:]))
    else:
        for one_item in item.as_list():
            if first_key in one_item:
                result.append(one_item[first_key])
        
    return result
        
    
    


from collections import defaultdict


class ItemTrack(object):    

    def __init__(self, item):
        self.item = item
        self.models = []
        self.fkeys = defaultdict(list)
        
        # grabbing `None` fkeys
        for fkey in item.relations:
            if fkey in item and item[fkey] is None:
                self.fkeys[fkey].append(None)


class ProcessKeeper(defaultdict):
    
    def __init__(self, item_structure=None):
        super().__init__(list)
        
        if item_structure:
            for item_cls, items in item_structure.items():
                for item in items:
                    self[item_cls].append(ItemTrack(item))
        
    def is_empty(self):
        for value in self.values():
            if value:
                return False
        return True


""" This module contains exceptions for
:py:class:`~save_to_db.core.item.Item` class configuration. These
exceptions are only raised when an item configuration is correct in its
structure but cannot be used, for example, when a nonexistent relation is used.
If, for example, a boolean value was used instead of a list, general exception
will be raise when trying to deal with a boolean as with a list.
"""

class ItemConfigError(Exception):
    """ General exception for configuration of
    :py:class:`~save_to_db.core.item.Item` class.
    """

class InvalidFieldName(ItemConfigError):
    """ Raised when configuration uses model field name that has double
    underscores in it.
    """

class RelatedItemNotFound(ItemConfigError):
    """ Raised when:
    
        - A related item was specified using a string path but no item
          class was found using the path.
        - During auto configuration of
          :py:class:`~save_to_db.core.item.Item` class relations, when no
          item was found for a specific model class.
    """

class MultipleRelatedItemsFound(ItemConfigError):
    """ Raised when:
    
        - A related item was specified using a string path with an
          item class name only and there are no such class in the
          referencing item
          module but more then one in other modules.
        - During auto configuration of
          :py:class:`~save_to_db.core.item.Item` relations, when more then one
          item was found for a specific model class.
    """

class ItemAdapterNotFound(ItemConfigError):
    """ Raised when an adapter for a particular item class was not found
    during class setup.
    """

class NonExistentRelationDefined(ItemConfigError):
    """ Raised when relation is set for a field that does not exist or is not
    a foreign key field in corresponding model.
    """

class NonExistentFieldUsed(ItemConfigError):
    """ Raised when `getters`, `creators`, or `nullables` or any other
    configuration value contains field name that `model_cls` does not have.
    """

class FieldsFromRelationNotAllowed(ItemConfigError):
    """ Raised when configuration variable contains
    fields referencing fields of a relations.
    """

class NorewriteRelationException(ItemConfigError):
    """ Raised when no-rewrite for a relationship set on both sides with
    different values.
    """

class NorewriteKeyUsedTwice(ItemConfigError):
    """ Raised when same no-rewrite key used twice (trough tuples).
    """
    
#--- ModelDeleter --------------------------------------------------------------

class DeleterSelectorsOrKeepersEmpty(ItemConfigError):
    """ Raised when trying to use ModelDelete with either `selector_fields`
    or `keeper_fields` empty.
    """
    
class DeleterXToManyRelationUsed(ItemConfigError):
    """ Raised when trying to use x-to-many relationship as selector or
    keeper field.
    """
    
class UnrefNonXToManyFieldUsed(ItemConfigError):
    """ Raise when trying to use not x-to-many field as a key of
    `unref_x_to_many` dictionary of item configuration.
    """


import pickle
import struct
from .item_cls_manager import item_cls_manager
from .merge_policy import MergePolicy
from .utils.item_collection import ItemCollection



class Persister(object):
    """ This class is used to persist items to database or save and load them
    from files.
    
    :param db_adapter: Instance of a subclass of
        :py:class:`~save_to_db.adapters.utils.adapter_base.AdapterBase` used
        to deal with items and ORM models.
    :param autocommit: If `True` commits changes to database each time an
        item is persisted.
    """
    
    def __init__(self, db_adapter, autocommit=False):
        self.db_adapter = db_adapter
        self.autocommit = autocommit
    
    
    #--- database adapter facade -----------------------------------------------
    
    def persist(self, item, commit=None):
        """ Saves item data into a database by creating or update appropriate
        database records.
        
        :param item: an instance of :py:class:`~.item_base.ItemBase` to persist.
        :param commit: If `True` commits changes to database. If `None` then
            `autocommit` value (initially set at creation time) is used.
            
        :returns: Item list and corresponding list of ORM  model lists.
        """
        result = self.db_adapter.persist(item)
        self.__commit_if_needed(commit)
        return result
    
    
    def merge_models(self, models, commit=None,
                     merge_policy=None, ignore_fields=None):
        """ Calls
        :py:meth:`~save_to_db.adapters.utils.adapter_base.AdapterBase.merge_models`
        method of
        :py:class:`~save_to_db.adapters.utils.adapter_base.AdapterBase`
        internal instance.
        
        :param commit: If `True` commits changes to database. If `None` then
            `autocommit` value (initially set at creation time) is used.
        :returns: First model from `models` into which all other models merged.
        """
        result_model = self.db_adapter.merge_models(
            models, merge_policy=merge_policy, ignore_fields=ignore_fields)
        self.__commit_if_needed(commit)
        return result_model
    
        
    def __commit_if_needed(self, commit):
        try:
            if commit or (commit is None and self.autocommit):
                self.commit()
        except:
            if commit or (commit is None and self.autocommit):
                self.rollback()
            raise
        

    def commit(self):
        """ Commits persisted items to database. """
        self.db_adapter.commit()
        
    
    def rollback(self):
        """ Rolls back current transaction. """
        self.db_adapter.rollback()
        
    
    def pprint(self, *models):
        """ Pretty prints `model` to console.
        
        :param \*models: List of models to print.
        """
        self.db_adapter.pprint(*models)

    
    #--- interface for working with files --------------------------------------
    
    def dumps(self, item):
        """ Converts an item into bytes.
        
        :param item: An instance of :py:class:`~.item_base.ItemBase` class.
        :returns: Encoded item as `bytes`.
        """
        item.process()
        model_cls = item.model_cls
        return pickle.dumps({
            'table_fullname': self.db_adapter.get_table_fullname(model_cls),
            'is_bulk_item': item.is_bulk_item(),
            'dict_wrapper': item.to_dict(),
            'collection_id': item.get_collection_id(),
        })
    
    
    def loads(self, data):
        """ Decodes `bytes` data into an instance of
        :py:class:`~.item_base.ItemBase`.
        
        :param data: Encoded item as `bytes`.
        :returns: An instance of :py:class:`~.item_base.ItemBase`.
        """
        result_data = pickle.loads(data)
        
        model_cls = self.db_adapter.get_model_cls_by_table_fullname(
            result_data['table_fullname'])
        item_cls = item_cls_manager.get_by_model_cls(
            model_cls, collection_id=result_data['collection_id'])[0]
        
        item = item_cls.Bulk() if result_data['is_bulk_item'] else item_cls()
        item = item.load_dict(result_data['dict_wrapper'])
        
        return item
    
        
    def dump(self, item, fp):
        """ Saves an item into a file.
        
        .. note::
            This method also saves the size of encoded item. So it is possible
            to save multiple items one after another into the same file and
            load them later.
        
        :param item: An item to be saved.
        :param fp: File-like object to save `item` into.
        """
        item_data = self.dumps(item)
        fp.write(struct.pack('>I', len(item_data)))
        fp.write(item_data)
        
    
    def load(self, fp):
        """ Loads and decodes one item from a file-like object.
        
        :param fp: File-like object to read from.
        :return: One item read from `fp` or `None` if there are no data to read
            anymore.
        """
        int_as_bytes = fp.read(4)
        if not int_as_bytes:
            return None
        l = struct.unpack('>I', int_as_bytes)[0]
        return self.loads(fp.read(l))
    
    #---model deleter ----------------------------------------------------------

    def execute_deleter(self, item_cls, commit=None):
        """ Deletes models according to `deleter_selectors`
        and `deleter_keepers`. See their description in
        :py:class:`~.item.Item` configuration.
        
        :param item_cls: :py:class:`~.item.Item` instance for which deletion
            must be executed.
        :param commit: If `True` commits changes to database. If `None` then
            `autocommit` value (initially set at creation time) is used.
        """
        if not item_cls.metadata['model_deleter']:
            return
        item_cls.metadata['model_deleter'].execute_delete(self.db_adapter)
        self.__commit_if_needed(commit)
    
    
    def execute_scope_deleter(self, collection_id, commit=None):
        """ Calls :py:meth:`~execute_deleter` method for all item classes in
        scope.
        
        :param collection_id: An ID of an
            :py:class:`~utils.item_collection.ItemCollection`
            (:py:class:`~.scope.Scope` is a subclass of it).
        :param commit: If `True` commits changes to database. If `None` then
            `autocommit` value (initially set at creation time) is used.
        """
        collection = ItemCollection.get_collection_by_id(collection_id)
        
        for item_cls in collection.get_all_item_classes():
            self.execute_deleter(item_cls, commit=False)
        
        self.__commit_if_needed(commit)
        
    #---merge policy -----------------------------------------------------------
    
    def create_merge_policy(self, policy, defaults=None):
        """ Creates an instance of :py:class:`~.merge_policy.MergePolicy`
        class.
        
        :param policy: *policy* argument for
            :py:class:`~.merge_policy.MergePolicy` class constructor.
        :param defaults: *defaults* argument for
            :py:class:`~.merge_policy.MergePolicy` class constructor.
        :returns: An instance of :py:class:`~.merge_policy.MergePolicy` class.
        """
        return MergePolicy(self.db_adapter, policy, defaults)
    
    
    


import os
import sys

DEBUG = True

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

SAVE_TO_BD_DIR = BASE_DIR
while not SAVE_TO_BD_DIR.endswith('save_to_db'):
    SAVE_TO_BD_DIR = os.path.dirname(SAVE_TO_BD_DIR)
SAVE_TO_BD_DIR = os.path.dirname(SAVE_TO_BD_DIR)
sys.path.insert(0, SAVE_TO_BD_DIR)

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
        #'NAME': os.path.join(BASE_DIR, 'db_django.sqlite3'),
    }
}

INSTALLED_APPS = ('orm_only',)

SECRET_KEY = 'SECRET_KEY'
